package clarion.acs;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.net.*;
import clarion.acs.net.auxiliary.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;
import clarion.nacs.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the ACS subsystem of CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void observeState( Feature[] sInput ): observes and forms the
 * input to ACS.
 * <li> public void reasoning(): run the learning/reasoning processes in ACS.
 * <li> public void update( Feature[] sInput, double[] reinforcements )
 * update the ACS subsystem.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements ACS subsystem and on the top of ACS implementation
 * structure and all the other classes relevant to ACS is declared inside it.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>Clarion</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class ACS {

   /** global settings. */
   protected Global  global;
   /** task-specific clarion settings. */
   protected TaskClarion taskClarion;
   /** eligible conditions of the external networks. */
   protected short[][][][] eligCondions;
   /** eligibility check  of the external networks. */
   protected EligibilityCheck eligCheck;
   /** EM is on or not. */
   protected boolean emOn;
   /** the probability of using EM to train the ACS. */
   protected double  emTrainAcsProb;
   /** AEM is on or not. */
   protected boolean aemOn;
   /** the probability of using AEM to train the ACS. */
   protected double  aemTrainAcsProb;
   /** current sensory input. */
   protected Feature sensoryInput[];
   /** desired outputs. */
   protected Feature desiredOutput[][];
   // variables for ACS components.
   /** the number of networks. */
   protected int netNum;
   /** the number of external networks. */
   protected int exNetNum;
   /** the number of active networks. */
   protected int activeNetNum;
   /** index to external network, WM network or goal structure network. */
   protected int exNetIdx, wmNetIdx, gsNetIdx;
   /** goal structure option. */
   protected int gsOption;
   /** WM is on or not. */
   protected boolean wmOnFlag;
   /** the active networks. */
   protected short[] activeNetworks[];
   /** the ACS networks. */
   protected AcsNet  nets[];
   /** current states of each network. */
   protected double[] netStates[];
   /** coordinating multiple external action net. */
   protected int  corExtMode;
   /** # of different acitons: GS, WM, orvaried EX. */
   protected int  actionTypeNum;
   /** knowledge types: BL, RER, IRL, FR. used for stochastic mode. */
   protected int  chosenType;
   /** current chosen actions from each active network.*/
   protected short  chosenActions[][][][];
   /** retrieval results from NACS. */
   protected GKSChunk retrievalResults[];

   /** construcutor.
    *  @param g the global setting. */
   public ACS( Global g )
   {
       global = g;
       global.setACS( this );
       taskClarion = global.getTaskClarion();
       // initialize eligibility of external networks.
       exNetNum = global.getExNetNum();
       eligCondions = new short[exNetNum][][][];
       for( int i = 0; i < exNetNum; i++ )
            eligCondions[i] = taskClarion.getEligibility(i);
       eligCheck = new EligibilityCheck( global, eligCondions );
       emOn = global.isEMon();
       if( emOn )
           emTrainAcsProb = global.getEmTrainAcsProb();
       aemOn = global.isAEMon();
       if( aemOn )
           aemTrainAcsProb = global.getAemTrainAcsProb();
       // initialize overall nets.
       netNum = global.getAcsNetNum();
       wmOnFlag = global.isWmOn();
       gsOption = global.getGsOption();
       wmNetIdx = global.getWmNetIdx();
       gsNetIdx = global.getGsNetIdx();
       nets = new AcsNet[ netNum ];
       corExtMode = global.getCorMultiExtMode();
       if( corExtMode == Global.COR_EXT_ALL ){
           activeNetNum = netNum;
           activeNetworks = new short[ activeNetNum ][ 2 ];
           for( int i = 0; i < exNetNum; i++ ){
                activeNetworks[ i ][ 0 ] = (short)Global.EX_TYPE;
                activeNetworks[ i ][ 1 ] = (short)i;
           }
           if( wmOnFlag ){
               activeNetworks[ wmNetIdx ][ 0 ] = (short)Global.WM_TYPE;
               activeNetworks[ wmNetIdx ][ 1 ] =(short)wmNetIdx;
           }
           if( gsOption != global.GS_NON ){
               activeNetworks[ gsNetIdx ][ 0 ] = (short)Global.GS_TYPE;
               activeNetworks[ gsNetIdx ][ 1 ] = (short)gsNetIdx;
           }
       }
       else   // SINGLE_MODE.
       {
           activeNetNum = 1;
           if( wmOnFlag )
               activeNetNum++;
           if( gsOption != global.GS_NON )
               activeNetNum++;
           activeNetworks = new short[ activeNetNum ][ 2 ];
           activeNetworks[ 0 ][ 0 ] = (short)Global.EX_TYPE;
           activeNetworks[ 0 ][ 1 ] = 0;
           if( wmOnFlag ){
               activeNetworks[ 1 ][ 0 ] = (short)Global.WM_TYPE;
               activeNetworks[ 1 ][ 1 ] = (short)wmNetIdx;
           }
           if( gsOption != global.GS_NON ){
               if( wmOnFlag ){
                   activeNetworks[ 2 ][ 0 ] = (short)Global.GS_TYPE;
                   activeNetworks[ 2 ][ 1 ] = (short)gsNetIdx;
               }
               else{
                   activeNetworks[ 1 ][ 0 ] = (short)Global.GS_TYPE;
                   activeNetworks[ 1 ][ 1 ] = (short)gsNetIdx;
               }
           }
       }
       //chosenActions = new short[activeNetNum][2][][];
       chosenActions = new short[netNum][2][][];
       netStates = new double[ netNum ][];
       for( int i = 0; i < netNum; i++ )
            netStates[ i ] = new double[ global.getIdnInputNum(i) ];
       // initialize External component.
       for( int i = 0; i < exNetNum; i++ )
            nets[ i ] = new EXNet( i, global );
       // initialize WM component if exist.
       if( wmOnFlag )
           nets[ wmNetIdx ] = new WMNet( wmNetIdx, global );
       // initialize GS component if exist.
       if( gsOption != global.GS_NON )  // a goal structure is selected.
           nets[ gsNetIdx ] = new GSNet( gsNetIdx, global );
   }

  public boolean isEligibleExNet( int netIdx )
  {
     for( int i = 0; i < activeNetworks.length; i++ )
          if( activeNetworks[i][0] == (short)Global.EX_TYPE && activeNetworks[i][1] == netIdx )
              return true;
     return false;
  }

  /** Returns the length of sensory input.
   *  @return the length of sensory input. */
   public int getSensoryLength()
   {
      return global.getSensoryInputLength();
   }

  /** Returns the length of goal input.
   *  @return the length of goal input. */
   public int getGsLength()
   {
      if( global.getGsOption() != Global.GS_NON )
          return global.getIdnInputNum( gsNetIdx );
      return 0;
   }

  /** Returns the length of WM input.
   *  @return the length of WM input. */
   public int getWmLength()
   {
      if( global.isWmOn() )
          return global.getIdnInputNum( wmNetIdx );
      return 0;
   }

  /** Returns the RER rules of the specific network in terms of strings.
   *  @param netIdx the index to the specific network.
   *  @return the RER rules of the specific network in terms of strings. */
   public String[] getRerRules( int netIdx )
   {
      if( nets[ netIdx ] != null )
          return nets[ netIdx ].getRerRules();
      else
          return (String[])null;
   }

  /** Returns the IRL rules of the specific network in terms of strings.
   *  @param netIdx the index to the specific network.
   *  @return the IRL rules of the specific network in terms of strings. */
   public String[] getIrlRules( int netIdx )
   {
      if( nets[ netIdx ] != null )
          return nets[ netIdx ].getIrlRules();
      else
          return (String[])null;
   }

  /** Returns the FR rules of the specific network in terms of strings.
   *  @param netIdx the index to the specific network.
   *  @return the FR rules of the specific network in terms of strings. */
   public String[] getFixRules( int netIdx )
   {
      if( nets[ netIdx ] != null )
          return nets[ netIdx ].getFixRules();
      else
          return (String[])null;
   }

  /** Returns the chosen actions by all networks.
   *  The chosen actions are represented by four dimensional array.
   *  The first dimension: the index to an network;
   *  The second dimension: NORMAL or NACS-control action;
   *  The third dimension: an index to a specific action dimension;
   *  The fourth dimesion: the active values in that action dimension.
   *  @returns the chosen actions. */
   public short[][][][] getChosenActions()
   {
       return CommonMethods.copy( chosenActions );
   }

  /** Returns the conclusions from all of the active networks in terms of chunks.
   *  The conclusions are represented by three dimension array.
   *  The first dimension: the index to an active network;
   *  The second dimension: an index to a specific knowledge type: IDN, RER, IRL, FR;
   *  The third dimesion: the conclusions from the knowledge type.
   *  @return the conclusions from all of the active networks in terms of chunks. */
   public Chunk[][][] getAllConclusions()
   {
      Chunk[][][] chunks = new Chunk[activeNetNum][][];
      for( int i = 0; i < activeNetNum; i++ )
           chunks[i] = nets[ activeNetworks[i][1] ].getAllConclusions();
      return chunks;
   }

   /** Returns the response time without NACS involved.
    *  @return the response time. */
   public int getRT()
   {
      int RT = 0;
      for( int i = 0; i < activeNetNum; i++ )
           RT = Math.max( RT, nets[ activeNetworks[i][1] ].getRT() );
      return RT;
   }

   /** Returns the response time with NACS involved.
    *  @param nacsRT the response time from NACS.
    *  @return the response time. */
   public int getRT( int nacsRT )
   {
      int RT = 0;
      for( int i = 0; i < activeNetNum; i++ )
           RT = Math.max( RT, nets[ activeNetworks[i][1] ].getRT( nacsRT ) );
      return RT;
   }

   /** Sets the sensory input.
    *  @param input the sensory input to set. */
   public void setSensoryInput( Feature[] input )
   {
      sensoryInput = input;
   }

   /** Sets the desired output.
    *  @param output the desired output to set. */
   public void setCurDesiredOutput( Feature[][] output )
   {
      desiredOutput = output;
   }

   /** Sets the retrieval results.
    *  @param result the result to set. */
   public void setRetrievalResult( GKSChunk[] result )
   {
      retrievalResults = result;
      if( wmOnFlag )
          ((WMNet)nets[ wmNetIdx ]).setRetrievalResult( retrievalResults );
   }

   /** Sets the current time.
    *  @param curTime current time. */
   public void setCurrentTime( long curTime )
   {
      for( int i = 0; i < nets.length; i++ )
           nets[ i ].setCurrentTime( curTime );
   }

   /** observe new sensory input, new GS and WM states if exists and deliver them
    *  to each active network and form new state including the three types of inputs
    *  if exists.
    *  @param sInput the sensory input. */
   public void observeState( Feature[] sInput )
   {
      // receive sensory input.
      setSensoryInput( sInput );
      // check which external net will be used.
      if( corExtMode == Global.COR_EXT_ONE ){
          exNetIdx = eligCheck.getEligibleNetwork( sInput );
          activeNetworks[ 0 ][ 1 ] = (short)exNetIdx;
      }
      // deliver sensory, GS and WM input to each eligible net if exists.
      for( int i = 0; i < activeNetNum; i++ ){
           nets[ activeNetworks[i][1] ].setSensoryInput( sensoryInput );
           if( activeNetworks[ i ][ 0 ] == Global.GS_TYPE ){
               nets[ activeNetworks[i][1] ].getGsState( netStates[ activeNetworks[i][1] ] );
               for( int j = 0; j < activeNetNum; j++ )
                    nets[ activeNetworks[j][1] ].setGsState( netStates[ activeNetworks[i][1] ] );
           }
           if( activeNetworks[ i ][ 0 ] == Global.WM_TYPE ){
               nets[ activeNetworks[i][1] ].getWmState( netStates[ activeNetworks[i][1] ] );
               for( int j = 0; j < activeNetNum; j++ )
                    nets[ activeNetworks[j][1] ].setWmState( netStates[ activeNetworks[i][1] ] );
           }
      }
      // form current state.
      for( int i = 0; i < activeNetNum; i++ )
          nets[ activeNetworks[i][1] ].formState();
   }

   /** Reasoning process in ACS. Let each active network decide actions and then
    *  get them. This method is called after the new state (input), including EX,
    *  GS, WM if exists, is given.
    */
   public void reasoning()
   {
      // let each active component reasoning.
      for( int i = 0; i < activeNetNum; i++ ){
           nets[ activeNetworks[i][1] ].reasoning();
           chosenActions[ activeNetworks[i][1] ] = nets[ activeNetworks[i][1] ].getChosenFullAction();
      }
   }

   /** Performs WM action. */
   public void performWmAction()
   {
      nets[ wmNetIdx ].perform();
   }

   /** Performs GS action. */
   public void performGsAction()
   {
      nets[ gsNetIdx ].perform();
   }

   /** Updates each active network. update GS, WM states if exists by performing
    *  the chosen actions. the EX action is already performed before this method
    *  is called.
    *  @param sInput new sensory input by performing the chosen external action.
    *  @param reinforcements the reinforcements received at this step.  */
   public void update( Feature[] sInput, double[] reinforcements )
   {
      // let each acs component update.
      for( int i = 0; i < activeNetNum; i++ ){
           int netIdx = activeNetworks[i][1];
           if( global.getIdnLearningOption( netIdx ) == Global.FIX )
               continue;
           if( activeNetworks[i][0] == global.EX_TYPE && global.getIdnLearningOption( netIdx ) == global.BP ){
               if( desiredOutput == null || desiredOutput[netIdx] == null )
                   System.out.println( "No desired output to train the IDN! (in ACS.java)" );
               else
                   nets[ netIdx ].setCurDesiredOutput( desiredOutput[netIdx] );
           }
           if( global.getIdnLearningOption( netIdx ) == global.QL )
                observeState( sInput );
           nets[ netIdx ].update( reinforcements[ netIdx ] );
      }
      // offline training ACS usning EM.
      if( emOn && Math.random() < emTrainAcsProb )
          offlineTrainByEm();
      // offline training ACS usning AEM.
      if( aemOn && Math.random() < aemTrainAcsProb )
          offlineTrainByAem();
   }

/******************************* PRIVATE METHOD *******************************/
  /** offline trains ACS by EM. */
  private void offlineTrainByEm()
  {
      for( int i = 0; i < activeNetNum; i++ ){
           //System.out.println( "the net = " + activeNetworks[i][1] );

           nets[ activeNetworks[i][1] ].offlineTrainByEm();
      }
  }

  /** offline trains ACS by AEM. */
  private void offlineTrainByAem()
  {
      for( int i = 0; i < activeNetNum; i++ )
           nets[ activeNetworks[i][1] ].offlineTrainByAem();
  }
}
