package clarion.acs.net;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.coordinate.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an ACS action-decision network. This class extends
 * the more generic class <em>GenrealNet</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void decideAction( int compIdx ): decide an action.
 * <li> public void reasoning(): the reasoning routine defined in this network.
 * <li> public void update( double reward ): update the network.
 * <li> public void formState(): compose current state given external, GS and WM states.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is to represent one ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>ACS</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class AcsNet extends GeneralNet {

  /** network type : EX, GS or WM. */
  protected int netType;
  /** the # of NACS-control output. */
  protected int fullControlOutputNum;
  /** the # of NACS-control output dims. */
  protected int controlOutputDimNum;
 /** the dim-val info of NACS-control outputs. */
  protected short[] fullControlOutputDVs;
  /** D-V info of bottom (IDN) outputs. */
  protected short[] fullOutputDVs;

  /** constructor.
   *  @param netIdx the network index.
   *  @param glboal the associated global settings. */
  public AcsNet( int netIdx, Global global )
  {
     super( Global.ACS );
     this.netIdx = netIdx;
     this.global = global;
     taskClarion = global.getTaskClarion();
     COMP_NUM = global.getAcsKnowledgeNum();
     netCompFlags = new boolean[ COMP_NUM ];
     netComps = new AcsNetComp[ COMP_NUM ];
     // initialize states.
     if( global.isExNet(netIdx) )
         exInputNum = global.getIdnInputNum( netIdx );
     else
         exInputNum = global.getSensoryInputLength();
     exState = new double[ exInputNum ];
     if( global.getGsOption() != Global.GS_NON ){
         int gsNetIdx = global.getGsNetIdx();
         //gsInputNum = global.getIdnInputNum( gsNetIdx );
         gsInputNum = global.getGsInputLength();
         gsState = new double[ gsInputNum ];
     }
     if( global.isWmOn() ){
         int wmNetIdx = global.getWmNetIdx();
         //wmInputNum = global.getIdnInputNum( wmNetIdx );
         wmInputNum = global.getWmInputLength();
         wmState = new double[ wmInputNum ];
     }
     inputNum = exInputNum + gsInputNum + wmInputNum;

     //System.out.println( "the input num = " + inputNum );

     state = new double[ inputNum ];
     // initialize IDN.
     //netCompFlags[ global.BL ] = global.isIdnOn( netIdx );
     netCompFlags[ global.BL ] = true;
     if( netCompFlags[global.BL] )
         netComps[ global.BL ] = new IDN( Global.ACS, inputNum, netIdx, global );
     // initialize ARS.
     netCompFlags[global.RER] = global.isRerOn( netIdx );
     if( netCompFlags[global.RER] )
         netComps[ global.RER ] = new RerRuleSet( Global.ACS, netIdx, global );
     netCompFlags[global.IRL] = global.isIrlOn( netIdx );
     if( netCompFlags[global.IRL] )
         netComps[ global.IRL ] = new IrlRuleSet( Global.ACS, netIdx, global );
     netCompFlags[global.FR] = global.isFrOn( netIdx );
     if( netCompFlags[global.FR] )
         netComps[ global.FR ] = new FixRuleSet( Global.ACS, netIdx, global );
     // for external output, actural output is the subset of full output,
     // don't use actural output as full output; for other types, use it.
     fullOutputNum = global.getAcsPhysicOutputNum( netIdx );
     fullOutputDimNum = global.getAcsPhysicOutputDimNum( netIdx );
     fullOutputDVs = global.getAcsPhysicOutputDimVals( netIdx );
     controlOutputDimNum = Global.ACT_DIM_NUM_NACS_CONTROL;
     fullControlOutputNum = global.getNacsControlPhysicOutputNum();
     fullControlOutputDVs = global.getNacsControlPhysicOutputDimVals();
     // initialize action activations  of each component.
     chosenFullAction = new short[ 2 ][][];
     chosenFullAction[Global.ACTION_NORMAL] = new short[fullOutputDimNum][];
     if( global.isNacsOn() )
         chosenFullAction[Global.ACTION_CONTROL] = new short[controlOutputDimNum][];
     outputActivations = new double[ COMP_NUM ][ 2 ][];
     for( int i = 0; i < COMP_NUM; i++ )
          outputActivations[i][Global.ACTION_NORMAL] = new double [ fullOutputNum ];
     if( global.isNacsOn() )
         for( int i = 0; i < COMP_NUM; i++ )
              outputActivations[i][Global.ACTION_CONTROL] = new double [ fullControlOutputNum ];
     // initialize the VCG sets.
     compVcgSets = new VCGSet[ COMP_NUM ][ 2 ][];
     initNormalVCGSets();
     if( global.isNacsOn() )
         initControlVCGSets();
     // initialize variables for combining components.
     combMode1 = global.getAcsCrossCombineMode(netIdx, 0 );
     combMode2 = global.getAcsCrossCombineMode(netIdx, 1 );
     succRates = new double[ COMP_NUM ];
     // initialize the combiner.
     double[] paras = global.getAcsCrossCombineParas( netIdx );
     if( combMode1 == global.STOCH_COMB_MODE ){
         stochComb = new StochasticComb( combMode2, COMP_NUM, paras, global );
         if( combMode2 == global.VAR_COMB_MODE ){
             c3s = global.getAcsCrossCombineVarParaAbove(netIdx);
             c4s = global.getAcsCrossCombineVarParaUnder(netIdx);
         }
     }
     else{ // bottom-up or top-guide.
         if( combMode2 < global.CT_COMB_MODE )
             wsComb = new WeightSumComb( combMode2, paras, global.getOverallActionSelectionTemp(netIdx) );
         else
             ctComb = new CorrectionComb();
     }
  }

  public boolean isChosenActionEmpty()
  {
     for( int i = 0; i < chosenFullAction.length; i++ )
          if( chosenFullAction[i] != null )
              for( int j = 0; j < chosenFullAction[i].length; j++ )
                   if( chosenFullAction[i][j] != null && chosenFullAction[i][j].length > 0 )
                       return false;
     return true;
  }

  /** Returns all conclusions by each knowledge type in terms of chunks.
   *  @return all conclusions by each knowledge type. */
  public Chunk[][] getAllConclusions()
  {
     Chunk[][] chunks = new Chunk[COMP_NUM][];
     for( int i = 0; i < COMP_NUM; i++ ){
          if( netCompFlags[i] )
              chunks[i] = ((AcsNetComp)netComps[i]).getAllConclusions();
     }
     return chunks;
  }

  /** Perform action. need be overrided by GS, WM net. */
  public void perform()
  {
  }

  /** Updates each component with the reinforcement at this step.
   *  @param reward the reinforcement received at this step. */
  public void update( double reward )
  {
     if( isChosenActionEmpty() )
         return;
     reinforcement = reward;
     // update IDN component.
     if( global.getIdnLearningOption( netIdx ) == global.QL )
         netComps[ global.BL ].setStates( state, exState, gsState, wmState );
     netComps[ global.BL ].update( reinforcement );
     // update other components.
     for( int i = 1; i < COMP_NUM; i++ ){
         if( netCompFlags[ i ] ){
             double threshold = 0.0;
             switch(i) {
               case Global.RER:
                    threshold = global.getRerPosThreshold( netIdx );
                    break;
               case Global.IRL:
                    threshold = global.getIrlPosThreshold( netIdx );
                    break;
             }
             boolean posFlag = ((IDN)netComps[ global.BL ]).getPositivity( i, threshold );
             netComps[ i ].update( posFlag );
         }
     }
  }

  /** weight sum combination method override the super method. */
   protected void weightSumCombine()
   {
      super.weightSumCombine();
      fillInVcgSets( Global.ACTION_CONTROL );
      wsComb.setVCGSets( controlVcgSets );
      chosenFullAction[Global.ACTION_CONTROL] = wsComb.combination();
   }

  /** Forms current state including EX, GS, WM if exists. */
  public void formState()
  {
     // form external state.
     if( global.isExNet( netIdx ) ){
         short exIdnInputDVs[];
         short exIdnInputFeatureInfo[][];
         exIdnInputDVs = global.getIdnInputDimVals( netIdx );
         exIdnInputFeatureInfo = global.getIdnInputFeatureInfo( netIdx );
         for( int i = 0; i < sensoryInput.length; i++ ){
              int idx = searchFeature( sensoryInput[ i ].getType(), sensoryInput[ i ].getOrder(),
                                       exIdnInputFeatureInfo );
              if( idx != -1 ){ // the feature exists in the input to this net.
                  int offset = 0;
                  for( int j = 0; j < idx; j++ )
                       offset += exIdnInputDVs[ j ];
                  short[] content = sensoryInput[ i ].getContent();
                  for( int j = 0; j < exIdnInputDVs[ idx ]; j++ )
                       exState[ offset + j ] = content[ j ];
              }
              else{
                  System.out.println( "not an eligible dim in AcsNet.java!" );
              }
         }
     }
     else{
         int offset = 0;
         for( int i = 0; i < sensoryInput.length; i++ ){
              short[] content = sensoryInput[ i ].getContent();
              for( int j = 0; j < content.length; j++ )
                   exState[ offset + j ] = content[ j ];
              offset += content.length;
         }
     }
     // form state including EX, GS, WM.
     for( int i = 0; i < exInputNum; i++ )
          state[i] = exState[i];
     int offset = exInputNum;
     if( global.getGsOption() != Global.GS_NON ){
         for( int i = 0; i < gsInputNum; i++ )
              state[offset+i] = gsState[i];
         offset += gsInputNum;
     }
     if( global.isWmOn() ){
         for( int i = 0; i < wmInputNum; i++ )
              state[offset+i] = wmState[i];
     }
  }

  /** offline train this network by EM. */
  public void offlineTrainByEm()
  {
     //((IDN)netComps[global.BL]).offlineTrainByEm();
     int emIterNum = global.getEmTrainAcsIterNum();
     NACS nacs = global.getNACS();
     EMAcsSample[] emSamples = nacs.getEmAcsSamples( netIdx, global.ACS );
     int emTrainSampleNum = emSamples.length;
     Vector sands = new Vector( emTrainSampleNum * emIterNum );
     for( int j = 0; j < emIterNum; j++ )
       for( int k = 0; k < emTrainSampleNum; k++ )
            sands.addElement( new Integer(k) );
     while( !sands.isEmpty() ){
       // decompose the step chunk.
       int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
       Chunk curState = emSamples[idx].getCurState();
       Chunk curAction = emSamples[idx].getCurAction();
       Chunk nextState = emSamples[idx].getNextState();
       double reward = emSamples[idx].getReward();
       // convert the state chunk into input to each component of the net.
       setStates( curState );
       // reasoning.
       reasoning();
       // deliver chosen action.
       for( int i = 0; i < fullOutputDimNum; i++ ){
            short[] values = curAction.getFeature(i).getValues();
            for( int j = 0; j < values.length; j++ )
                 chosenFullAction[Global.ACTION_NORMAL][i][j] = values[j];
       }
       deliverAction();
       // convert the nextState chunk into input to each component of the net.
       if( nextState != null )
           setStates( nextState );
       // update each component
       update( reward );
    }
  }

  /** offline train this network by AEM. */
  public void offlineTrainByAem()
  {
     //((IDN)netComps[global.BL]).offlineTrainByAem();
     int aemIterNum = global.getAemTrainAcsIterNum();
     NACS nacs = global.getNACS();
     AbsEMSample[] aemSamples = nacs.getAemSamples( netIdx );
     int aemTrainSampleNum = aemSamples.length;
     Vector sands = new Vector( aemTrainSampleNum * aemIterNum );
     for( int j = 0; j < aemIterNum; j++ )
       for( int k = 0; k < aemTrainSampleNum; k++ )
            sands.addElement( new Integer(k) );
     while( !sands.isEmpty() ){
       int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
       Chunk curState = aemSamples[idx].getCurState();
       Chunk curAction = aemSamples[idx].getCurAction();
       Chunk nextState = aemSamples[idx].getNextState();
       double reward = aemSamples[idx].getReward();
       double[] stateDistribution = aemSamples[idx].getStateDistributions();
       Chunk[] nextStates = aemSamples[idx].getPotentialNextStates();
       // convert the state chunk into input to each component of the net.
       setStates( curState );
       // reasoning.
       reasoning();
       // deliver chosen action.
       for( int i = 0; i < fullOutputDimNum; i++ ){
            short[] values = curAction.getFeature(i).getValues();
            for( int j = 0; j < values.length; j++ )
                 chosenFullAction[Global.ACTION_NORMAL][i][j] = values[j];
       }
       deliverAction();
       // convert the nextState chunk into input to each component of the net.
       if( nextState != null )
           setStates( nextState );
       // update each component
       update( reward, nextStates, stateDistribution );
     }
  }

/************************** private methods ***********************************/
  /** search feature. */
  private int searchFeature( int type,  int order, short[][] info )
  {
      int idx = -1;
      for( int i = 0; i < info.length; i++ ){
           if( info[i][0] == type && info[i][1] == order ){
               idx = i;
               break;
           }
      }
      return idx;
  }

  /** used for offline training using AEM. */
  private void update( double reward, Chunk[] nextStates, double[] stateDistribution )
  {
     if( isChosenActionEmpty() )
         return;
       reinforcement = reward;
     // update IDN component.
     if( global.getIdnLearningOption( netIdx ) == global.QL )
         netComps[ global.BL ].setStates( state, exState, gsState, wmState );
     ((IDN)netComps[ global.BL ]).update( reinforcement, nextStates, stateDistribution );
     // update other components.
     for( int i = 1; i < COMP_NUM; i++ ){
         if( netCompFlags[ i ] ){
             double threshold = 0.0;
             switch(i) {
               case 1:
                    threshold = global.getRerPosThreshold( netIdx );
                    break;
               case 2:
                    threshold = global.getIrlPosThreshold( netIdx );
                    break;
               case 3:
                    threshold = reward;
                    //threshold = 0.1;
                    break;
             }
             boolean posFlag = ((IDN)netComps[ global.BL ]).getPositivity( i, threshold );
             netComps[ i ].update( posFlag );
         }
     }
  }


  /** initialize the VCG sets. */
  private void initNormalVCGSets()
  {
     for( int i = 0; i < COMP_NUM; i++ ){
       compVcgSets[i][Global.ACTION_NORMAL] = new VCGSet[ fullOutputDimNum ];
       for( int j = 0; j < fullOutputDimNum; j++ ){
            int type = global.getOutputMultiValueProperty( netType )[j];
            if( type == Global.SINGLE_ACTIVE_VALUED ){
                int vcgNum = 2;
                short[] elemNums = new short[vcgNum];
                elemNums[0] = (short)(fullOutputDVs[j] - 2);
                elemNums[1] = 2;
                compVcgSets[i][Global.ACTION_NORMAL][j] = new VCGSet( type, vcgNum, elemNums );
            }
            if( type == Global.MULTI_ACTIVE_VALUED ){
                int vcgNum = (short)fullOutputDVs[j] / 2;
                short[] elemNums = new short[vcgNum];
                for( int k = 0; k < vcgNum; k++ )
                     elemNums[k] = 2;
                compVcgSets[i][Global.ACTION_NORMAL][j] = new VCGSet( type, vcgNum, elemNums );
            }
       }
     }
     normalVcgSets = new VCGSet[ COMP_NUM ][];
     for( int i = 0; i < COMP_NUM; i++ )
          normalVcgSets[i] = compVcgSets[i][Global.ACTION_NORMAL];
  }

  /** initialize the VCG sets. */
  private void initControlVCGSets()
  {
    for( int i = 0; i < COMP_NUM; i++ ){
       compVcgSets[i][Global.ACTION_CONTROL] = new VCGSet[ controlOutputDimNum ];
       for( int j = 0; j < controlOutputDimNum; j++ ){
            int type = global.getOutputMultiValueProperty( Global.CONTROL_TYPE )[j];
            if( type == Global.SINGLE_ACTIVE_VALUED ){
                int vcgNum = 2;
                short[] elemNums = new short[vcgNum];
                elemNums[0] = (short)(fullControlOutputDVs[j] - 2);
                elemNums[1] = 2;
                compVcgSets[i][Global.ACTION_CONTROL][j] = new VCGSet( type, vcgNum, elemNums );
            }
            if( type == Global.MULTI_ACTIVE_VALUED ){
                int vcgNum = fullControlOutputDVs[j] / 2;
                short[] elemNums = new short[vcgNum];
                for( int k = 0; k < vcgNum; k++ )
                     elemNums[k] = 2;
                compVcgSets[i][Global.ACTION_CONTROL][j] = new VCGSet( type, vcgNum, elemNums );
            }
       }
    }
    controlVcgSets = new VCGSet[ COMP_NUM ][];
    for( int i = 0; i < COMP_NUM; i++ )
         controlVcgSets[i] = compVcgSets[i][Global.ACTION_CONTROL];
  }
}
