package clarion.acs.net;
import clarion.acs.net.auxiliary.*;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.net.auxiliary.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an IRL (Independent Rule Learning) rule. It extends
 * the more generic rule class <em>TLRule</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean decideAction( short[] state, short[] exState, short[] gsState,
 * short[] wmState ): Decides an action by this IRL rule.
 * <li> public void generalize(): the generalization routine.
 * <li> public void specialize(): the specialization routine.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the IRL rule which is one of the three rule knowledge
 * types in the top level of an ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>TLRuleSet</em> object.
 * @author Xi Zhang
 *
 */

public class IrlRule extends TLRule {

   final static double LOG_2 = Math.log( 2 );
   /** the group the agent is in, decide which IRL subset is used. */
   protected int groupId;
   /** the rule index. */
   protected int ruleIdx;
   /** parameters for computing Information Gain. */
   protected double c5, c6;
   /** threshold for deletion. */
   protected double threshold4;
   /** threshold for generalization. */
   protected double threshold6;
   /** threshold for specialization. */
   protected double threshold7;
   /** the IRL rule form. */
   protected IrlRuleForm ruleForm;

   /** constructor.
    *  @param groupIdx the index to the group the associated agent is in.
    *  @param netIdx the network index.
    *  @param setIdx the rule subset index to the rule set.
    *  @param ruleIdx the rule index to the rule subset.
    *  @param global the associated global settings. */
   public IrlRule( int groupIdx, int subsysIdx, int netIdx, int setIdx, int ruleIdx, Global global )
   {
       super( subsysIdx, netIdx, Global.IRL, setIdx, global );
       this.groupId = groupIdx;
       this.ruleIdx = ruleIdx;
       c5 = global.getIrlIGParaAbove( netIdx );
       c6 = global.getIrlIGParaUnder( netIdx );
       threshold4 = global.getIrlDeleteThreshold( netIdx );
       threshold6 = global.getIrlGenThreshold( netIdx );
       threshold7 = global.getIrlSpecThreshold( netIdx );
       ruleForm = global.getTaskClarion().getIrlRule( groupIdx, netIdx, setIdx, ruleIdx );
   }

   /** Return if the rule is useles or not. */
   public boolean isUseless()
   {
       double pm = attributes.getPM();
       double nm = attributes.getNM();
       /*
       double ig = Math.log((pm+c5)/(pm+nm+c6))/LOG_2;
       if( ig < threshold4 )
           return true;
       else
           return false;
       */
       //return taskClarion.getIrlIG( netIdx, pm, nm, threshold4 ) < 0.0;
       boolean flag = taskClarion.getIrlIG( netIdx, pm, nm, threshold4 ) < 0.0;
       //if( flag ){
       //    System.out.println( "match rate = " + (pm+1)/(pm+nm+2) );
       //}

       return flag;
   }

  /** Decides the action.
    *  @param state curent state.
    *  @param exState external state.
    *  @param gsState goal structure state.
    *  @param wmState working memory state.
    *  @return <tt>true</tt> if this rule is matched.
    *  <tt>false>, otherwise. */
   public boolean decideAction( short[] state, short[] exState, short[] gsState, short[] wmState )
   {
       ruleSupport = ruleForm.determineAction( state, exState, gsState, wmState, suggestedAction );
       if( Math.abs( ruleSupport - 1.0 ) < 0.01 )
           return true;
       else
           return false;
   }

   /** Returns if this rule can be generalized. */
   public boolean canGeneralize()
   {
       double pm = attributes.getPM();
       double nm = attributes.getNM();
       return taskClarion.getIrlIG( netIdx, pm, nm, threshold6 ) >= 0.0;
   }

   /** Returns if this rule can be specialized. */
   public boolean canSpecialize()
   {
       double pm = attributes.getPM();
       double nm = attributes.getNM();
       return taskClarion.getIrlIG( netIdx, pm, nm, threshold7 ) < 0.0;
   }

   /** the generalization routine. */
   public void generalize()
   {
       ruleForm.generalize();
   }

   /** the specialization routine. */
   public void specialize()
   {
       ruleForm.specialize();
   }

/*************************** OVERRIDED METHODS *********************************/
   /** methods from class <em>Object</em> */
   public String toString()
   {
       String str = "IRL RULE( " + netIdx + ", " + setIdx + ", " + ruleIdx + " ) = \n";
       double pm = attributes.getPM();
       str += "pm = " + String.valueOf( pm );
       double nm = attributes.getNM();
       str += ", nm = " + String.valueOf( nm );
       double rate = (pm+c5)/(pm+nm+c6);
       str += ", match rate = " + String.valueOf( rate );
       str += "\n";
       return str;
  }
}