/********************* IMPLEMENTATION OF RER RULE ***************************/
package clarion.acs.net;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.Iterator;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;
import clarion.acs.net.auxiliary.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the RER (Rule Extraction Refinement) Rule set. It
 * extends the more generic rule set class <em>TLRuleSet</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void update( boolean successFlag ): Updates with flag indicating
 * current step is successful or not.
 * <li> public void reasoning(): the reasoning routine.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the RER rules which is one of the three rule knowledge
 * types in the top level of an ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>AcsANet</em> object.
 * @author Xi Zhang
 *
 */

public class RerRuleSet extends TLRuleSet {
/************************** Variables of RerRuleSet ******************************/
   /** probability of rule extraction. */
   protected double P_re;
   /** generalization threshold.  */
   protected double threshold1;
   /** specalizaiton threshold. */
   protected double threshold2;
   /** merge frequency. */
   protected int mergePerSteps;
   /** steps since the last app of the rule. */
   protected int timeSteps;
   /** link used for managing the rule set. */
   protected HashSet ruleLink;
   /** match-all rules. */
   protected TreeMap matchAllRuleSet = new TreeMap();

/************************* begin of RerRuleSet methods  **************************/
  /** constructor.
   *  @param netIdx network index.
   *  @param glboal the associated global settings. */
  public RerRuleSet ( int subsysIdx, int netIdx, Global global )
  {
       super( subsysIdx, netIdx, Global.RER, global );
       MAX_AGE = (int)( 1 / global.getRerDensity( netIdx ) );
       genMode = global.getRerGenSpecValueRange( netIdx );
       P_re = global.getRerExtractionProb( netIdx );
       threshold1 = global.getRerGenThreshold( netIdx );
       threshold2 = global.getRerSpecThreshold( netIdx );
       mergePerSteps = global.getRuleMergeInterval( netIdx );
       ruleLink = new HashSet();
  }

  /** Returns the RER rules in terms of strings.
   *  @return the RER rules.  */
  public String[] getRules()
  {
     ruleReps[ ALL ] = "THE RULE SET: \n";
     Iterator iter = ruleLink.iterator();
     while( iter.hasNext() )
            ruleReps[ ALL ] += iter.next();
     return ruleReps;
  }

  /** Returns the active RER rules in terms of string.
   *  @return the active RER rules.  */
  public String getActiveRules()
  {
     String str = "CURRENT ACTIVE RULE SET: \n";
     Iterator iter = ruleLink.iterator();
     while( iter.hasNext() )
            str += iter.next();
     return str;
  }

  /** Returns the number of positive matches.
   *  @return the number of positive matches.  */
  public double getPM()
  {
     double pm = 0.0;
     Iterator iter = ruleLink.iterator();
     while( iter.hasNext() ){
         ActiveRule rule = (ActiveRule)iter.next();
         pm += rule.getOverallPM();
     }
     return pm;
  }

  /** Returns the number of negative matches.
   *  @return the number of negative matches.  */
  public double getNM()
  {
     double nm = 0.0;
     Iterator iter = ruleLink.iterator();
     while( iter.hasNext() ){
         ActiveRule rule = (ActiveRule)iter.next();
         nm += rule.getOverallNM();
     }
     return nm;
  }

  /** The reasoning routine. */
  public void reasoning()
  {
      RULE_NUM = getActiveRuleNum();
      ruleUtls = new double[ RULE_NUM ];
      recommendRules = new short[ RULE_NUM ];
      ruleList = new RerRule[ RULE_NUM ];
      Iterator iter = ruleLink.iterator();
      int idx = 0;
      while( iter.hasNext() )
          ruleList[idx++] = ((ActiveRule)iter.next()).getRule();

      //System.out.println( "the # of active rules = " + RULE_NUM );
      //for( int i = 0; i < ruleList.length; i++ )
      //    System.out.println( ruleList[i] );

      super.reasoning();
  }

  /** Updates current RER rule set at each trial with flag indicating current step
   *  is successful or not. Now recommend rules only from active rules.
   *  @param successFlag the flag indicating current step is successful or not. */
  public void update( boolean successFlag )
  {
      updateMatchAllRuleSet( chosenAction, successFlag );
      /* search all the matching rules both from the active and inactive nodes.
         and update all the statistics of each matching rule. */
      int counter = 0;
      //ruleReps[ ALL ] = "All RER rules: \n";
      ruleReps[ MATCH_BOTH ] = "Match both condition and action RER rules: \n";
      boolean extRuleFlag, genRuleFlag, specRuleFlag;
      extRuleFlag = genRuleFlag = specRuleFlag = false;
      Iterator iter = ruleLink.iterator();
      while( iter.hasNext() ){
           ActiveRule activeRule = ((ActiveRule)iter.next());
           RerRule rule = activeRule.getRule();
           //ruleReps[ ALL ] += "( active rule )" + rule;
           rule.matchAction( chosenAction );
           rule.update( successFlag, currentTime );
           Object[] children = activeRule.getChildren();
           if( rule.isBasicMatched() ){
               counter++;
               ruleReps[ MATCH_BOTH ] += rule;
               updateActionUsage();
               for( int i = 0; i < children.length; i++ ){
                    RerRule child = (RerRule)children[i];
                    //ruleReps[ ALL ] += "( ------- inactive rule )" + child;
                    child.decideAction( state, exState, gsState, wmState );
                    child.matchAction( chosenAction );
                    child.update( successFlag, currentTime );
                    if( child.isBasicMatched() )
                        counter++;
               }
           }
           else{
               /*
               for( int i = 0; i < children.length; i++ ){
                    RerRule child = (RerRule)children[i];
                    ruleReps[ ALL ] += "( ------- inactive rule )" + child;
               }
               */
           }
      }
      RerRule extRule = null;
      if( counter == 0 ){   // no rule covers this step in the top level.
          //if( successFlag && Math.random() < P_re ){  // extract a new rule.
          if( successFlag ){  // extract a new rule.
              if( genMode == Global.GS_CONTI_VAL || genMode == Global.GS_NCONT_VAL )
                  extRule = new RerRule( subsysIdx, netIdx, global, state, chosenAction, false );
              else
                  extRule = new RerRuleEx( subsysIdx, netIdx, global, state, chosenAction, false, INPUT_DIM_NUM );
              extraction( extRule );
              extRuleFlag = true;
          }
      }
      else{  //genealization and specialization only for the active rules
         ActiveRule[] genRules = new ActiveRule[ ruleLink.size() ];
         ActiveRule[] specRules = new ActiveRule[ ruleLink.size() ];
         int count = 0;
         iter = ruleLink.iterator();
         while( iter.hasNext() ){
           ActiveRule activeRule = ((ActiveRule)iter.next());
           RerRule rule = activeRule.getRule();
           if( rule.isBasicMatched() )
               continue;
           boolean genFlag = rule.canGeneralized();
           boolean specFlag = rule.canSpecialized();
           double ig1 = infoGain(rule,(RerRule)matchAllRuleSet.get( new KeyToValue(
                        OUTPUT_NUM, CONTROL_OUTPUT_NUM, chosenAction, outputOffsets, controlOutputOffsets )));
           if( successFlag && genFlag ){
               double ig2 = maxInfoGain( rule, Global.GENERALIZATION );
               if( ( ig1 > threshold1 ) && ( ig2 >= 0 ) ){
                   //generalization( activeRule );
                   genRules[count++] = activeRule;
                   genRuleFlag = true;
               }
           }
           if( !successFlag && specFlag ){
               double ig3 = maxInfoGain( rule, Global.SPECIFICATION );
               if( ( ig1 < threshold2 ) && ( ig3 > 0 ) ){
                   //specialization( activeRule );
                   specRules[count++] = activeRule;
                   specRuleFlag = true;
               }
           }
         }
         if( genRuleFlag ){
             for( int i = 0; i < count; i++ )
                  generalization( genRules[i] );
         }
         if( specRuleFlag ){
             for( int i = 0; i < count; i++ )
                  specialization( specRules[i] );
         }
      }

      //ruleReps[ ALL ] += "extRuleFlag = " + extRuleFlag + "\n";
      //ruleReps[ ALL ] += "genRuleFlag = " + genRuleFlag + "\n";
      //ruleReps[ ALL ] += "specRuleFlag = " + specRuleFlag + "\n";
      //if( extRuleFlag )
      //    ruleReps[ ALL ] += "the extracted rule = " + extRule;

      handleUselessRule();
      mergeActiveRule( extRuleFlag, genRuleFlag, specRuleFlag );
      discountRuleStatistics();

      //ruleReps[ ALL ] += "after merge active rule \n";
  }

  /*
  public void deleteRule( short[] input, short[] action )
  {
      Iterator iter = ruleLink.iterator();
      while( iter.hasNext() ){
          RerRule rule = ((ActiveRule)iter.next()).getRule();
          rule.decideAction( state, exState, gsState, wmState );
          rule.matchAction( action );
          if( !rule.isBasicMatched() )
              iter.remove();
      }
  }
  */

/*************************** OVERRIED METHODS *********************************/
  /** methods from class <em>Object</em> */
  public String toString()
  {
     String str = null;
     Iterator iter = ruleLink.iterator();
     while( iter.hasNext() )
        str += ((ActiveRule)iter.next());
     return str;
  }

/*************************** PRIVATE METHODS **********************************/
  /** Returns current active rule numbers. */
  private int getActiveRuleNum()
  {
      return ruleLink.size();
  }

  /** Updates the match-all rules. Match-all rule means that the condition of
      the rule matches all the possible states. */
  private void updateMatchAllRuleSet( short[][][] action, boolean succFlag )
  {
      /*
      // convert 3-dimensional array to 2-dimensional array.
      short[][] action;
      if( !global.isNacsOn() || subsysIdx == Global.SUPER )
          action = act[Global.ACTION_NORMAL];
      else{
          action = new short[ act[Global.ACTION_NORMAL].length + act[Global.ACTION_CONTROL].length ][];
          CommonMethods.copy( action, act );
      }
      */

      short[] cond = new short[ INPUT_NUM ];
      for( int i = 0; i < INPUT_NUM; i++ )
           cond[ i ] = 1;
      RerRule rule = null;
      if( genMode == Global.GS_CONTI_VAL || genMode == Global.GS_NCONT_VAL )
          rule = new RerRule( subsysIdx, netIdx, global, cond, action, true );
      else
          rule = new RerRuleEx( subsysIdx, netIdx, global, cond, action, true, INPUT_DIM_NUM );
      KeyToValue key = new KeyToValue( OUTPUT_NUM, CONTROL_OUTPUT_NUM, chosenAction,
                                       outputOffsets, controlOutputOffsets );
      if( !matchAllRuleSet.containsKey( key ) )
          matchAllRuleSet.put( key, rule );
      ((RerRule)matchAllRuleSet.get( key )).updateMatchAll( succFlag );
  }

   /** Return the information gain. the measure for goodness of a rule. */
   private double infoGain( RerRule ruleA, RerRule ruleB )
   {
      double PMa = ruleA.getPM();
      double NMa = ruleA.getNM();
      double PMb = ruleB.getPM();
      double NMb = ruleB.getNM();
      return taskClarion.getRerIG( netIdx, PMa, NMa, PMb, NMb );
   }

   /** Returns maxIG(C',C). */
   private double maxInfoGain( RerRule rule, int mode )
   {
      int[] retVal;
      double PMa = 0;
      double NMa = 0;
      double PMb = rule.getPM();
      double NMb = rule.getNM();
      if( genMode == Global.GS_CONTI_VAL || genMode == Global.GS_NCONT_VAL ){
          retVal = rule.searchMaxIGLoc( mode );
          PMa = rule.getMaxMatches( retVal[0], retVal[1], mode )[0];
          NMa = rule.getMaxMatches( retVal[0], retVal[1], mode )[1];
      }
      else{
          PMa = ((RerRuleEx)rule).getMaxMatches( mode )[0];
          NMa = ((RerRuleEx)rule).getMaxMatches( mode )[1];
      }
      return taskClarion.getRerIG( netIdx, PMa, NMa, PMb, NMb );
   }

   /** extract a new rule from the current state and the recommend action. */
   private void extraction( RerRule rule )
   {
      ActiveRule activeRule = new ActiveRule( rule, netIdx, taskClarion );
      ruleLink.add( activeRule );
   }

  /* search the oldest rule according tO the # of steps occurred since the */
  /* last  application of the rule and delete it */
  /*
   private void deleteOldestRule( )
   {
      int age, maxAge = 0;
      ActiveRule oldRule = null;
      Iterator iter = ruleLink.iterator();
      while( iter.hasNext() ){
           ActiveRule activeRule = ((ActiveRule)iter.next());
           RerRule rule = activeRule.getRule();
           age = rule.getAge();
           if( maxAge < age ){
               maxAge = age;
               oldRule = activeRule;
           }
      }
      iter = ruleLink.iterator();
      while( iter.hasNext() ){
           ActiveRule activeRule = ((ActiveRule)iter.next());
           RerRule rule = activeRule.getRule();
           age = rule.getAge();
           if( age == maxAge ){
               iter.remove();
               break;
           }
      }
   }
   */

   /** generalization amounts to adding an additional value to one input dimension
    *  in the condition of a rule, so that the rule will have more opportunities
    *  of matching inputs. */
   private boolean generalization( ActiveRule activeRule )
   {
       //ruleReps[ ALL ] += "before generalization, rule set = " + this.getActiveRules() + "\n";
       //ruleReps[ ALL ] += "before generalization, rule = " + activeRule + "\n";

       if( !ruleLink.contains( activeRule ) ) // as child of one active rule.
           return false;

       RerRule rule = activeRule.getRule();
       rule.setGenerailizedFlag( true );
       int[] loc  = new int[2];
       if( genMode == Global.GS_CONTI_VAL || genMode == Global.GS_NCONT_VAL ){
           loc = rule.getMaxIGLoc( 0 );
           rule.addCondition( loc[0], loc[1] );
       }
       else
           ( (RerRuleEx)rule ).generalizeADim();
       coverActiveRules( activeRule );
       rule.resetStatistics();

       //ruleReps[ ALL ] += "after generalization, rule set = " + this.getActiveRules() + "\n";
       //ruleReps[ ALL ] += "after generalization, rule = " + activeRule + "\n";

       return true;
   }

   private void coverActiveRules( ActiveRule activeRule )
   {
      int size = ruleLink.size();
      ActiveRule[] children = new ActiveRule[ size ];
      Object[] ruleArr = ruleLink.toArray();
      int idx = 0;
      int count = 0;
      while( idx < size ){
         ActiveRule otherRule = (ActiveRule)ruleArr[idx];
         int retV1 = activeRule.cover( otherRule );
         int retV2 = otherRule.cover( activeRule );
         if( retV1 == -2 && retV2 != -2 ) // one-direction covers one rule.
             children[count++] = otherRule;
         idx++;
      }
      if( count > 0 ){
          ruleLink.remove( activeRule );
          for( int i = 0; i < count; i++ ){
               activeRule.addChildren( children[i] );
               ruleLink.remove( children[i] );
          }
          ruleLink.add( activeRule );
      }
   }

   /** specialization amounts to removing one value from one input dimension in
    *  the condition of a rule, so that the rule will have less opportunities
    *  of matching inputs. */
   private boolean specialization( ActiveRule activeRule )
   {
      if( !ruleLink.contains( activeRule ) ) // as child of one active rule.
          return false;
      RerRule rule = activeRule.getRule();
      rule.setSpecializedFlag( true );
      int[] loc = new int[2];
      if( genMode == Global.GS_CONTI_VAL || genMode == Global.GS_NCONT_VAL ){
          loc = rule.getMaxIGLoc( 1 );
          rule.delCondition( loc[0], loc[1] );
      }
      else
          ( (RerRuleEx)rule ).specializeADim();
      restoreChildren( activeRule );
      rule.resetStatistics();
      return true;
   }

  /** restore those rules in the children list of the original rule that are not
   *  covered by the specialized rule and the other active rules. */
   private void restoreChildren( ActiveRule parent )
   {
      RerRule[] restoredChildren = parent.removeChildren();
      for( int i = 0; i < restoredChildren.length; i++ ){
           ActiveRule activeRule = new ActiveRule( restoredChildren[i], netIdx, taskClarion );
           ruleLink.add( activeRule );
      }
   }

   /** search if any useless rule exists according to the age of the rule and delete it. */
   private void handleUselessRule()
   {
      Iterator iter = ruleLink.iterator();
      while( iter.hasNext() ){
         ActiveRule activeRule = (ActiveRule)(iter.next());
         int age = activeRule.getRule().getAge();
         if( age >= MAX_AGE ){
             iter.remove();

             //System.out.println( "delete rule happens this step!" );

         }
      }
  }

  /** merge rules according to what the tutorial says. */
  private void mergeActiveRule( boolean extRuleFlag, boolean genRuleFlag, boolean specRuleFlag )
  {
       if( (++timeSteps)<mergePerSteps )
           return;
       if( !extRuleFlag && !genRuleFlag && !specRuleFlag )
           return;
       timeSteps = 0;

       //ruleReps[ ALL ] += "Merge rule happens this step! \n";

       int size = ruleLink.size();
       //Object[] ruleArr1 = ruleLink.toArray();
       //Object[] ruleArr2 = ruleLink.toArray();
       int idx1, idx2;
       idx1 = idx2 = 0;
       while( idx1 < ruleLink.size() && idx2 < ruleLink.size() ){
           ActiveRule activeRule1 = ( (ActiveRule)ruleLink.toArray()[idx1] );
           RerRule rule1 = activeRule1.getRule();
           ActiveRule activeRule2 = ( (ActiveRule)ruleLink.toArray()[idx2] );
           RerRule rule2 = activeRule2.getRule();
           if( idx1 == idx2 ){
               idx2 = (++idx2) % size;
               idx1 = ( idx2 == 0 ) ? idx1 + 1 : idx1;
               continue;
           }
           int retV1 = rule1.cover( rule2 );
           int retV2 = rule2.cover( rule1 );
           if( retV1 == -1 && retV2 == -1 || retV1 == -1 && retV2 >= 0 ||
               retV1 >= 0 && retV2 == -1 ){  // not covered at all.
               idx2 = (++idx2) % size;
               idx1 = ( idx2 == 0 ) ? idx1 + 1 : idx1;
               continue;
           }
           boolean b1 = ruleLink.remove( activeRule1 );
           boolean b2 = ruleLink.remove( activeRule2 );
           if( retV1 == -2 || retV2 == -2 ){  // covered entirely
               if( retV1 == -2 ){
                   activeRule1.addChildren( activeRule2 );
                   ruleLink.add( activeRule1 );
               }
               else{
                   activeRule2.addChildren( activeRule1 );
                   ruleLink.add( activeRule2 );
               }
           }
           if( retV1 >= 0 && retV2 >= 0 ){ // merge.
                // create a new rule with the same statistics as the bigger rule.
                RerRule rule;
                if( genMode == Global.GS_CONTI_VAL || genMode == Global.GS_NCONT_VAL )
                    rule = (RerRule)rule1.clone();
                else
                    rule = (RerRuleEx)rule1.clone();
                short[] arr1 = rule1.getCondition( retV1 );
                short[] arr2 = rule2.getCondition( retV1 );
                for( int i = 0; i < inputDVs[ retV1 ]; i++ )
                     if( arr1[ i ] == 0 && arr2[ i ] == 1 )
                         rule.addCondition( retV1, i );
                ActiveRule activeRule = new ActiveRule( rule, netIdx, taskClarion );
                activeRule.addChildren( activeRule1 );
                activeRule.addChildren( activeRule2 );
                ruleLink.add( activeRule );
           } // end of if( retVal >=0 )
           idx1 = idx2 = 0;
        }  // end of while
   }

  /** update age which is used for deleting an oldest rule when rule number
      reaches  the max number, discount the statistics of all the rules. */
  private void discountRuleStatistics()
  {
      trialIdx++;
      // discount Match-all rules.
      Object[] matchAllRules = matchAllRuleSet.values().toArray();
      for( int i = 0; i < matchAllRules.length; i++ ){
           ((RerRule)matchAllRules[i]).aging();
           if( trialIdx >= episodeLength )
               ((RerRule)matchAllRules[i]).discountStatistics();
      }
      // discount RER rules
      Iterator iter = ruleLink.iterator();
      while( iter.hasNext() ){
           ActiveRule activeRule = ((ActiveRule)iter.next());
           RerRule rule = activeRule.getRule();
           rule.aging();
           if( trialIdx >= episodeLength )
               rule.discountStatistics();
           Object[] children = activeRule.getChildren();
           for( int i = 0; i < children.length; i++ ){
               ((RerRule)children[i]).aging();
               if( trialIdx >= episodeLength )
                   ((RerRule)children[i]).discountStatistics();
           }
      }
      if( trialIdx >= episodeLength )
          trialIdx = 0;
  }
}