package clarion.acs.net.auxiliary;
import clarion.system.*;
import clarion.common.*;

 /**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is used to check the eligibility of current input, to determine
   which network should be used in current situation.

   Each network has an eligible condition for current input to match.
   1. Such eligible condition has the format if(C1 or C2...or Cn) then use net j,
      ( each Ci should be exclusive of each other ).
   2. Ci has the format (ci1 and ci2... and cin).
   3. the above formats should be satisfied by user input before doing eligibility
      checking.

 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public int getEligibleNetwork( Feature[] input ): returns the eligible network
 * given current input.
 * <li> public EligibilityCheck( Global global, Feature[][][] eligs ): constructor.
 * <li> public EligibilityCheck( Global global, short[][][][] eligConds ): constructor.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the eligibility check of each ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>ACS</em> object.
 * @author Xi Zhang
 *
 */

public class EligibilityCheck{
  /** the number of networks. */
  protected int netNum;
  /** the global setttings. */
  protected Global global;
  /** eligibility of each network.
      dim1: network NO.
      dim2: "OR" conditions Ci in a network.
      dim3: "AND" condiitons Cij.
      @see clarion.system.TaskClarion#getEligibility(int netIdx) */
  protected Feature[][][] eligs;

  /** constructor.
   *  @param glboal the global settings.
   *  @param eligs the eligibility of ACS networks in terms of feature list.
   *  @see clarion.system.TaskClarion#getEligibility(int netIdx) */
  public EligibilityCheck( Global global, Feature[][][] eligs )
  {
     this.global = global;
     netNum = eligs.length;
     this.eligs = eligs;
  }

  /** constructor.
   *  @param glboal the global settings.
   *  @param eligs the eligibility of ACS networks in terms of four dimensional
   *  array.
   *  @see clarion.system.TaskClarion#getEligibility(int netIdx) */
  public EligibilityCheck( Global global, short[][][][] eligConds )
  {
     this.global = global;

     //System.out.println("in ELIG, global feat type = " + global.getIdnInputFeatureInfo(0)[0][0] );
     //System.out.println("in ELIG, global feat idx = " + global.getIdnInputFeatureInfo(0)[0][1] );

     netNum = eligConds.length;
     eligs = new Feature[netNum][][];
     for( int i = 0; i < netNum; i++ ){
       int orCondNum = eligConds[i].length;
       eligs[i] = new Feature[orCondNum][];
       for( int j = 0; j < orCondNum; j++ ){
          short[] featIndices = eligConds[i][j][0];
          int andCondNum = featIndices.length;
          eligs[i][j] = new Feature[andCondNum];
          for( int k = 0; k < andCondNum; k++ ){
               int idx = featIndices[k];

               //System.out.println("i = " + i + ", idx = " + idx );

               short[] featInfo = global.getIdnInputFeatureInfo(i)[idx];
               short[] con = eligConds[i][j][k+1];
               String label = global.getFeatureNames( featInfo[0], featInfo[1] );
               eligs[i][j][k] = new Feature( label, featInfo[0], featInfo[1], con, global );
          }
       }
     }
  }

  /** Returns the eligible network given current input.
   *  @param input current input.
   *  @return the index to eligible network. */
  public int getEligibleNetwork( Feature[] input )
  {
     int netIdx = 0;
     for( int i = 0; i < netNum; i++ )
      for( int j = 0; j < eligs[i].length; j++ )
      { // check if current input meets eligibility of a network.
           boolean eligFlag = true;
           for( int k = 0; k < eligs[i][j].length; k++ )
           {// check if current input meets one of the eligibilities of a network.
                for( int m = 0; m < input.length; m++ )
                     if( !eligs[i][j][k].isTheSameFeature( input[m] ) ||
                         !eligs[i][j][k].isContained( input[m] ) )
                         eligFlag = false;
                     else{
                         eligFlag = true;
                         break;
                     }
                if( !eligFlag )
                    break;
           }
           if( eligFlag )
           { // meets one of the eligibilities of a network.
               netIdx = i;
               return netIdx;
           }
      }
      return netIdx;
  }
}