package clarion.acs.net.auxiliary;
import java.util.*;
import java.util.TreeMap;
import clarion.acs.net.*;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is the base class to represent the generic rule knowledge type in top
 * level of ACS. It extends the class <em>AcsNetComp</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public short[][] decideAction(): returns the recommended action from
 * current rule set.
 * <li> public void reasoning(): the reasoning routine.
 * <li> public void update( boolean successFlag ): Updates with flag indicating
 * current step is successful or not.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is the basic representation of the generic rule knowledge type in
 * top level of a action-decision network in ACS. Concrete rule knoweledge types
 * (rule set) such as RER rule set, IRL rule set and FR rule set will extends it.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>AcsANet</em> object.
 * @author Xi Zhang
 *
 */

public class TLRuleSet extends AcsNetComp {
   protected int MATCH_CONDITION = 0, MATCH_BOTH = 1, ALL = 2;
   /** rule type: RER, IRL or FR. */
   protected int ruleType;
   /** rule list. */
   protected TLRule[] ruleList;
   /** rule # of current rule set, the length of ruleList. */
   protected int RULE_NUM;
   /** maximal number of steps withnot applying a rule. */
   protected int MAX_AGE;
   /** generalization mode. */
   protected int genMode;
   /** partial match flag. */
   protected boolean partialMatchFlag;
   /** threshold for partial match. */
   protected double  threshold_PM;
   /** utility flag. */
   protected boolean UTL_FLAG;
   /** rule utilities used for action decision. */
   protected double  ruleUtls[];
   /** the activations of the activated actions. */
   protected TreeMap activationList = new TreeMap();
   /** current state. */
   protected short[] state;
   /** current external state. */
   protected short[] exState;
   /** current goal state. */
   protected short[] gsState;
   /** current working memory state. */
   protected short[] wmState;
   /** used for MCS's minotoring conclusions. */
   protected int c20;
   /** results from Top Level. */
   protected Chunk[] tlResults;
   /** number of rule matched. */
   protected int ruleMatchNum;
   /** the recommended rule index to the rule list. */
   protected int recommendRule;
   /** the recommended rule indices to the rule list. */
   protected short[] recommendRules;
   /** rule representations : MATCH_CONDITION, MATCH_BOTH, ALL. */
   protected String ruleReps[];
   /** used for cross combination. */
   protected double  SUCC_C3, SUCC_C4;
   /** used for cross combination. */
   protected double  succRate, sumPM, sumNM;
   /** used for updating stats. */
   protected int episodeLength, trialIdx;
   /** action BLA option: constant or variable. used for RT. */
   protected int actionBLAoption;
   /** action usages info. */
   protected TreeMap actionUsageInfoTM = new TreeMap();

  /** constructor.
   *  @param netIdx network index.
   *  @param ruleType rule type.
   *  @param glboal the associated global settings. */
   public TLRuleSet (int subsysIdx, int netIdx, int ruleType, Global global )
   {
       this.subsysIdx = subsysIdx;
       this.netIdx = netIdx;
       this.ruleType = ruleType;
       this.global = global;
       taskClarion = global.getTaskClarion();
       INPUT_NUM = global.getARInputNum( netIdx, ruleType );
       OUTPUT_NUM = global.getAROutputNum( netIdx, ruleType );
       INPUT_DIM_NUM = global.getARInputDimNum( netIdx, ruleType );
       OUTPUT_DIM_NUM = global.getAROutputDimNum( netIdx, ruleType );
       inputDVs = global.getARInputDimVals( netIdx, ruleType );
       outputDVs = global.getAROutputDimVals( netIdx, ruleType );
       inputOffsets = new short[INPUT_DIM_NUM];
       outputOffsets = new short[OUTPUT_DIM_NUM];
       CommonMethods.calcOffsets( inputOffsets, inputDVs );
       CommonMethods.calcOffsets( outputOffsets, outputDVs );
       if( subsysIdx == Global.ACS ){
           netType = global.isExNet(netIdx) ? Global.EX_TYPE :(global.isWmNet(netIdx) ? Global.WM_TYPE : Global.GS_TYPE );
           FULL_OUTPUT_NUM = global.getAcsPhysicOutputNum( netIdx );
           FULL_OUTPUT_DIM_NUM = global.getAcsPhysicOutputDimNum( netIdx );
           fullOutputDVs = global.getAcsPhysicOutputDimVals( netIdx );
           fullOutputOffsets = new short[fullOutputDVs.length];
           CommonMethods.calcOffsets( fullOutputOffsets, fullOutputDVs );
       }
       if( subsysIdx == Global.SUPER ){
       }
       CONTROL_OUTPUT_NUM = global.getNacsControlOutputNum();
       CONTROL_OUTPUT_DIM_NUM = Global.ACT_DIM_NUM_NACS_CONTROL;
       controlOutputDVs = global.getNacsControlOutputDimVals();
       controlOutputOffsets = new short[CONTROL_OUTPUT_DIM_NUM];
       CommonMethods.calcOffsets( controlOutputOffsets, controlOutputDVs );
       FULL_CONTROL_OUTPUT_NUM = global.getNacsControlPhysicOutputNum();
       FULL_CONTROL_OUTPUT_DIM_NUM = Global.ACT_DIM_NUM_NACS_CONTROL;
       fullControlOutputDVs = global.getNacsControlPhysicOutputDimVals();
       fullControlOutputOffsets = new short[FULL_CONTROL_OUTPUT_DIM_NUM];
       CommonMethods.calcOffsets( fullControlOutputOffsets, fullControlOutputDVs );
       chosenAction = new short[ 2 ][][];
       chosenAction[Global.ACTION_NORMAL] = new short[ OUTPUT_DIM_NUM ][];
       if( global.isNacsOn() && subsysIdx == Global.ACS )
           chosenAction[Global.ACTION_CONTROL] = new short[ CONTROL_OUTPUT_DIM_NUM ][];
       outputActivations = new double[ 2 ][];
       outputActivations[Global.ACTION_NORMAL] = new double[ OUTPUT_NUM ];
       if( global.isNacsOn() && subsysIdx == Global.ACS )
           outputActivations[Global.ACTION_CONTROL] = new double[ CONTROL_OUTPUT_NUM ];
       // get user options.
       partialMatchFlag = global.isRulePartMatchOn( netIdx, ruleType, 0 );
       if( partialMatchFlag )
           threshold_PM = global.getPMThreshold( netIdx );
       UTL_FLAG = global.isRuleUtilityOn( netIdx, ruleType, 0 );
       ruleReps = new String[ 3 ];
       episodeLength = global.getEpisodeLength();
       // initialize the variables used for cross combination.
       if( global.getAcsCrossCombineMode( netIdx, 1 ) == Global.VAR_COMB_MODE ){
           SUCC_C3 = global.getAcsCrossCombineVarParaAbove( netIdx )[ruleType];
           SUCC_C4 = global.getAcsCrossCombineVarParaUnder( netIdx )[ruleType];
       }
       // initialize the action usages for calculating BLA if the bla option is set.
       if( global.isActionChunkBlaOn( netIdx, ruleType, 0 ) )
           actionBLAoption = global.getAcsChunkBlaOption( netIdx, ruleType, 0 );
       else
           actionBLAoption = Global.EMPTY;
       if( actionBLAoption == Global.VARIED_OPTION )
           actionUsageInfoTM = new TreeMap();  // action usages.
       // initialize the conclusions for monitoring.
       if( global.isMcsOn() )
           c20 = global.getMonitorAcsTopConcNum();
       // action decider.
       double temper = global.getARActionSelectionTemp( netIdx, ruleType );
       actionDecider = new StochasticDecider( temper );
   }

   /** Returns the rules in terms of strings.
    *  @return the rules. */
   public String[] getRules()
   {
      return ruleReps;
   }

   /** Returns the all rules in terms of strings..
    *  @return the all rules. */
   public String getAllRules()
   {
      return ruleReps[ ALL ];
   }

   /** Returns the only condition-matched rules (from the active rules) in terms of strings.
    *  @return the only condition-matched rules. */
   public String getMatchConditionRules()
   {
      return ruleReps[ MATCH_CONDITION ];
   }

   /** Returns the both matched rules (from the active rules) in terms of strings.
    *  @return the both matched rules. */
   public String getMatchRules()
   {
      return ruleReps[ MATCH_BOTH ];
   }

   /** Returns the positive matches.
    *  @return the positive matches. */
   public double getPM()
   {
      sumPM = 0.0;
      for( int i = 0; i < RULE_NUM; i++ )
           if( ruleList[ i ] != null )
               sumPM += ruleList[ i ].getPM();
      return sumPM;
   }

   /** Returns the negative matches.
    *  @return the negative matches. */
   public double getNM()
   {
      sumNM = 0.0;
      for( int i = 0; i < RULE_NUM; i++ )
           if( ruleList[ i ] != null )
               sumNM += ruleList[ i ].getNM();
      return sumNM;
   }

   /** Returns the success rate.
    *  @return the success rate. */
   public double getSuccRate()
   {
      getPM();
      getNM();
      succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );
      return succRate;
   }

   /** Converts outputActivations into its full-bottom format since current
    *  format is just a subset of the full-bottom format. And returns all
    *  the activated actions' activations with its full-bottom format.
    *  @returns the activations with its full-bottom format. */
   public double[][] getOutputActivations()
   {
       double[][] fullFormatActivations = new double[ 2 ][];
       fullFormatActivations[Global.ACTION_NORMAL] = new double[ FULL_OUTPUT_NUM ];
       if( global.isNacsOn() && subsysIdx == Global.ACS )
           fullFormatActivations[Global.ACTION_CONTROL] = new double[ FULL_CONTROL_OUTPUT_NUM ];
       int[] dimNums = { OUTPUT_DIM_NUM, CONTROL_OUTPUT_DIM_NUM };
       for( int j = 0; j < 2; j++ )
         if( fullFormatActivations[j] != null )
         for( int i = 0; i < dimNums[j]; i++ ){
            int idx = global.getAROutputFeatureInfo( netIdx, ruleType )[i][1];
            // don't need relocation.
            if( j == Global.ACTION_CONTROL || (subsysIdx == Global.ACS && netType != Global.EX_TYPE) )
                idx = i;
            int length = fullOutputDVs[idx];
            if( j == Global.ACTION_CONTROL )
                length = fullControlOutputDVs[idx];
            double[] activations = new double[length];
            int type = global.getOutputMultiValueProperty( netType )[idx];
            if( j == Global.ACTION_CONTROL )
                type = global.getOutputMultiValueProperty( Global.CONTROL_TYPE )[idx];
            if( type == Global.SINGLE_ACTIVE_VALUED ){
                if( j == Global.ACTION_NORMAL )
                    System.arraycopy( outputActivations[j], outputOffsets[i], activations, 0, length - 2 );
                else
                    System.arraycopy( outputActivations[j], controlOutputOffsets[i], activations, 0, length - 2 );
                // fill in SINGLE_ACTIVE_VALUED flag.
                activations[length-2] =  1;
                activations[length-1] =  0;
            }
            else{
                if( j == Global.ACTION_NORMAL )
                    for( int k = 0; k < outputDVs[i]; k++ )
                         activations[ 2 * k + Global.ELEM_ACTIVE ] = outputActivations[j][outputOffsets[i] + k];
                else
                    for( int k = 0; k < controlOutputDVs[i]; k++ )
                         activations[ 2 * k + Global.ELEM_ACTIVE ] = outputActivations[j][controlOutputOffsets[i] + k];
                // fill in MULTI_ACTIVE_VALUED flag.
                activations[length-2] =  0;
                activations[length-1] =  1;
            }
            if( j == Global.ACTION_NORMAL )
                System.arraycopy( activations, 0, fullFormatActivations[j], fullOutputOffsets[idx], length );
            else
                System.arraycopy( activations, 0, fullFormatActivations[j], fullControlOutputOffsets[idx], length );
       }
       return fullFormatActivations;
   }

   /** Returns all the activated actions' activations.
    *  @param arr the array to store the activations. */
   public void getOutputActivations( double[][] arr )
   {
      double[][] activations = getOutputActivations();
      for( int i = 0; i < arr.length; i++ )
           arr[ i ] = activations[ i ];
   }

   /** Returns all conclusions decided by the active rules.
    *  @return all conclusions. */
   public Chunk[] getAllConclusions()
   {
      tlResults = fillInFinalResults();
      return tlResults;
   }

   /** Returns response time given the actuation time of top level.
    *  @return response time. */
   public int getResponseTime( int TLAT )
   {
      calcRT( TLAT );
      return responseTime;
   }

   /** Sets the chosen action.
    *  @param action the action to fill in the chosen action.
    *  In new design, this parameter has full # of output dimensions.
    *  */
   public void setChosenAction( short[][][] fullAction )
   {
      if( subsysIdx == Global.ACS && !global.isExNet( netIdx ) ){
           chosenAction = CommonMethods.copy( fullAction );
           return;
      }
      // convert the action with fill format into the partial format.
      short[][] dimFeatInfo = global.getAROutputFeatureInfo( netIdx, ruleType );
      for( int i = 0; i <  OUTPUT_DIM_NUM; i++ ){
           int idx = dimFeatInfo[i][1];
            // don't need relocation.
            if( subsysIdx == Global.ACS && netType != Global.EX_TYPE )
                idx = i;
           chosenAction[Global.ACTION_NORMAL][i] = CommonMethods.copy(fullAction[Global.ACTION_NORMAL][idx]);
      }
      if( subsysIdx == Global.ACS && global.isNacsOn() )
          chosenAction[Global.ACTION_CONTROL] = CommonMethods.copy(fullAction[Global.ACTION_CONTROL]);
   }

   /** Sets states.
    *  @param state curent state.
    *  @param exState extenal state.
    *  @param gsState goal state.
    *  @param wmState working memory state. */
   public void setStates( double[] state, double[] exState, double[] gsState, double[] wmState )
   {
      if( this.state == null )
          this.state = new short[ state.length ];
      for( int i = 0; i < state.length; i++ )
           this.state[ i ] = (short)state[ i ];
      if( this.exState == null )
          this.exState = new short[ exState.length ];
      for( int i = 0; i < exState.length; i++ )
           this.exState[ i ] = (short)exState[ i ];
      if( this.gsState == null && gsState != null ){
          this.gsState = new short[ gsState.length ];
          for( int i = 0; i < gsState.length; i++ )
               this.gsState[ i ] = (short)gsState[ i ];
      }
      if( this.wmState == null && wmState != null ){
          this.wmState = new short[ wmState.length ];
          for( int i = 0; i < wmState.length; i++ )
               this.wmState[ i ] = (short)wmState[ i ];
      }
   }

   /** Sets current time.
    *  @param curTime the time to set. */
   public void setCurrentTime( int curTime )
   {
      currentTime = curTime;
   }

   /** Returns the recommended action from current rule set.
    *  @return the recommended action in its full size format
    *  covering all of the action dims. */
   public short[][][] decideAction()
   {
      if( ruleMatchNum == 0 )
          return null;
      int recommendIdx = 0;
      recommendIdx = actionDecider.getRecommendation( ruleUtls, ruleMatchNum );
      recommendRule = recommendRules[ recommendIdx ];
      suggestedAction = ruleList[recommendRule].getAction();
      if( subsysIdx == Global.ACS && !global.isExNet( netIdx ) ){
          fullSuggestedAction = CommonMethods.copy( suggestedAction );
          return fullSuggestedAction;
      }
      // convert the action with partial format into the full format.
      fullSuggestedAction = new short[2][][];
      fullSuggestedAction[Global.ACTION_CONTROL] = CommonMethods.copy( suggestedAction[Global.ACTION_CONTROL] );
      fullSuggestedAction[Global.ACTION_NORMAL] = new short[FULL_OUTPUT_DIM_NUM][];
      for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
           int idx = global.getAROutputFeatureInfo( netIdx, ruleType )[i][1];
           fullSuggestedAction[Global.ACTION_NORMAL][idx] = CommonMethods.copy(suggestedAction[Global.ACTION_NORMAL][i]);
      }
      return CommonMethods.copy( fullSuggestedAction );
   }

   /** the reasoning process. if utility is not on, use rule support to select
    *  action. */
   public void reasoning()
   {
      ruleMatchNum = 0;
      activationList.clear();
      CommonMethods.resetArray( outputActivations, 0.0 );
      for( int i = 0; i < RULE_NUM; i++ ){
           ruleUtls[ i ] = 0.0;
           recommendRules[ i ] = -1;
      }
      ruleReps[ ALL ] = "All rules: \n";
      ruleReps[ MATCH_CONDITION ] = "Match condition rules: \n";
      for( int i = 0; i < RULE_NUM; i++ ){
           if( ruleList[i] == null )
               continue;
           boolean matchFlag = ruleList[i].decideAction( state, exState, gsState, wmState );

           //System.out.println( "i = " + i + ", ruleList[i] = " + ruleList[i] );

           ruleReps[ ALL ] += ruleList[i];

           double support = ruleList[i].getRuleSupport();
           if( matchFlag || partialMatchFlag && support > threshold_PM ){
               ruleReps[ MATCH_CONDITION ] += ruleList[i];
               recommendRules[ ruleMatchNum ] = (short)i;
               if( UTL_FLAG )
                   ruleUtls[ ruleMatchNum ] = ruleList[i].getUtility();
               else
                   ruleUtls[ ruleMatchNum ] = support;
               ruleMatchNum++;
               updateActionActivations( ruleList[i] );
           }
           //updateActionActivations( ruleList[i] );
      }
      fillInOutputActivation();
   }

  /** Updates this rule set at each trial with flag indicating current step
   *  is successful or not. Now recommend rules only from active rules.
   *  @param successFlag the flag indicating current step is successful or not. */
   public void update( boolean successFlag )
   {
      ruleReps[ MATCH_BOTH ] = "Match both condition and action rules: \n";
      for( int i = 0; i < ruleMatchNum; i++ ){
           int idx = recommendRules[i];
           boolean actionMatchFlag = ruleList[ idx ].isActionMatched( chosenAction );
           if( actionMatchFlag ){
               ruleReps[ MATCH_BOTH ] += ruleList[ idx ];
               updateActionUsage();
           }
           ruleList[idx].update( actionMatchFlag, successFlag, currentTime );
      }
   }

  /** Updates this rule set at each trial with currently received reinforcement.
   *  @param reward currently received reinforcement. */
   public void update( double reward )
   {
   }

   /** Calculates the response time given the actuation time of top level.
    *  @param TLAT the actuation time of top level. */
   public void calcRT( int TLAT )
   {
      responseTime = 1500;
      if( actionBLAoption == Global.EMPTY || ruleMatchNum == 0 )
          return;
      double actionBla = calcChosenActionBLA();
      for( int i = 0; i < ruleMatchNum; i++ ){
           int idx = recommendRules[i];
           if( ruleList[ idx ].isActionMatched( chosenAction ) ){
               ruleList[idx].calcBLA( currentTime );
               ruleList[idx].calcTLDT( actionBla );
               ruleList[idx].calcRT( TLAT );
               responseTime = Math.max( responseTime, ruleList[idx].getResponseTime() );
           }
      }
  }

  /** Updates the action usage. */
  public void updateActionUsage()
  {
     if( actionBLAoption != Global.VARIED_OPTION )
           return;
     // update action BLAs.
     KeyToValue key = new KeyToValue( OUTPUT_NUM, CONTROL_OUTPUT_NUM, chosenAction,
                                      outputOffsets, controlOutputOffsets );
     if( !actionUsageInfoTM.containsKey( key ) )
          actionUsageInfoTM.put(key,new ActionUsageInfo(currentTime,global.getAcsChunkBlaParas(netIdx,ruleType,0)));
      ActionUsageInfo value = (ActionUsageInfo)actionUsageInfoTM.get(key);
      value.increaseUsage( currentTime );
  }

/****************************** PRIVATE METHODS ********************************/
  /** calculate the BLA of currently chosen action, used for RT. */
  private double calcChosenActionBLA()
  {
      if( actionBLAoption == global.CONSTANT_OPTION )
          return global.getAcsChunkBla( netIdx, ruleType, 0 );
      KeyToValue key = new KeyToValue( OUTPUT_NUM, CONTROL_OUTPUT_NUM, chosenAction,
                                       outputOffsets, controlOutputOffsets );
      if( !actionUsageInfoTM.containsKey( key ) )
          actionUsageInfoTM.put(key,new ActionUsageInfo(currentTime,global.getAcsChunkBlaParas(netIdx,ruleType,0)));
      ActionUsageInfo value = (ActionUsageInfo)actionUsageInfoTM.get(key);
      value.calcBla( currentTime );
      return value.getBla();
  }

  /** Updates action activations. */
  private void updateActionActivations( TLRule rule )
  {
     double support = rule.getRuleSupport();
     if( support < threshold_PM )
         return;
     short[][][] action = rule.getAction();
     if( action == null )
         return;
     KeyToValue key = new KeyToValue( OUTPUT_NUM, CONTROL_OUTPUT_NUM, action,
                                      outputOffsets, controlOutputOffsets );
     if( activationList.containsKey( key ) ){
         double value = ((Double)activationList.get( key )).doubleValue();
         value = Math.max( support, value );
         activationList.put(key, new Double(value));
     }
     else
         activationList.put(key, new Double(support));
  }

  /** fill in output activation. */
  private void fillInOutputActivation()
  {
     Object[] keys = activationList.keySet().toArray();
     Object[] values = activationList.values().toArray();
     int actionNum = keys.length;
     for( int i = 0; i < actionNum; i++ ){
          BitSet actionKey = ((KeyToValue)keys[i]).getKeys();
          for( int j = 0; j < OUTPUT_NUM; j++ ){
               if( actionKey.get(j) )
                   // average of the activation of a dimensional value from all of activated actions.
                   outputActivations[Global.ACTION_NORMAL][j] += (((Double)values[i]).doubleValue())/actionNum;
          }
          if( outputActivations[Global.ACTION_CONTROL] != null )
          for( int j = 0; j < CONTROL_OUTPUT_NUM; j++ ){
               if( actionKey.get( OUTPUT_NUM + j ) )
                   // average of the activation of a dimensional value from all of activated actions.
                   outputActivations[Global.ACTION_CONTROL][j] += (((Double)values[i]).doubleValue())/actionNum;
          }
     }
  }

  /** fill in the most activated conclusions from top level. */
  private Chunk[] fillInFinalResults()
  {
     int resultNum = ruleMatchNum;
     double[] strengths = new double[ resultNum ];
     for( int i = 0; i < resultNum; i++ )
          strengths[i] = ruleList[recommendRules[i]].getRuleSupport();
     SortableElement[] sortableElements = new SortableElement[ resultNum ];
     for( int i = 0; i < resultNum; i++ )
          sortableElements[i] = new SortableElement( strengths[i], recommendRules[i] );
     Arrays.sort( sortableElements );
     int c = Math.min( c20, resultNum );
     int len = (global.isMcsOn())? c : resultNum;
     Chunk[] results = new Chunk[ len ];
     try{
        for( int i = 0; i < len; i++ ){
             int idx = sortableElements[resultNum-i-1].getChunkIdx();
             results[i] = (Chunk)(ruleList[idx].getActionChunk().clone());
        }
     }
     catch( Exception e ){
            System.out.println( e + ", clone chunk error in TLRuleSet.java." );
     }
     return results;
  }
}