package clarion.common;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the algorithm for backpropagation network.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void forwardPass(): is to calculate the actual outputs given current
 * input.
 * <li> public void backwardPass(): is to update all of the weights inside the neural
 * network.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is the base class implementing the ACS implicit leanring processes.
 * Other classes will extend it to implement other implicit learning process such as
 * Q-learning, Simplified Q-learning.
 * <li> Usage of this class. <br>
 * This base class is considered as one of basic components in a general network
 * implemented by the class <em>GeneralNet</em> and is instantiated inside a concrete
 * class extending the class <em>GeneralNet</em>.
 * </ul>
 * @author Xi Zhang
 *
 */

public class BPNet{
    /** global settings. */
    protected Global global;
    /** the numbers of units in the layer of input, hidden and output. */
    protected int nInput, nHidden, nOutput;
    /** the input, hidden and output. */
    protected double[] Input,Output,Hidden;
    /** the momentum for hidden and output. */
    protected double[] OutputMomentum, HiddenMomentum;
    /** the thresholds of output and hidden. */
    protected double[] OutputThresholds, HiddenThresholds;
    /** the derivative of output and hidden. */
    protected double[] OutputDeriv,HiddenDeriv;
    /** the errors of output and hidden. */
    protected double[] Errors, HiddenErrors;
    /** the desired output. */
    protected double[] DesiredOutput;
    /** the weights. */
    protected double[][] InputToHiddenWeights, HiddenToOutputWeights;
    protected double[][] ItoHWeights, HtoOWeights;   // used for restoration.
    /** the momentums. */
    protected double[][] InputToHiddenMomentum, HiddenToOutputMomentum;
    //protected double  HINITWEIGHT = 0.05;    // brackets for the initial weights
    //protected double  LINITWEIGHT = -0.05;
    protected double  HINITWEIGHT = 0.01;    // brackets for the initial weights
    protected double  LINITWEIGHT = -0.01;
    //protected double  HINITTHRESHOLD = 1;    // brackets for the initial weights
    //protected double  LINITTHRESHOLD = -1;
    protected double  HINITTHRESHOLD = 0.1;    // brackets for the initial weights
    protected double  LINITTHRESHOLD = -0.1;
    /** momentum. used for update of weights. */
    protected double Momentum;     //
    /** learning rate. */
    protected double Eta;
    /** tolerance of error...sorta */
    protected double RZero;
    protected boolean debug;
    protected double sumSqErrors;
    /** variables for RT. */
    protected int    BLRT, BLPT, BLDT, BLAT;
    /** variables for success rate. */
    protected double PM, NM, succRate;
    /** parameters for success rate. */
    protected double SUCC_C3 = 1.0, SUCC_C4 = 2.0;
    /** variables for information of dimensional values. */
    protected short[] inputDVs, outputDVs;
    /** the start index of each action's dim in net output.  */
    protected short[] outputOffsets;

    // old constructor.
    public BPNet(int numInput, int numHidden, int numOutput )
    {
        // # of units in input, output, hidden layers
        nInput = numInput;
        nOutput = numOutput;
        nHidden = numHidden;
        // value of input, hidden and output vectors
        Input = new double[nInput];
        Hidden = new double[nHidden];
        Output = new double[nOutput];
        // weight vectors
        InputToHiddenWeights = new double[nInput][nHidden];
        HiddenToOutputWeights = new double[nHidden][nOutput];
        ItoHWeights = new double[nInput][nHidden];
        HtoOWeights = new double[nHidden][nOutput];
        // momentum vectors
        InputToHiddenMomentum = new double[nInput][nHidden];
        HiddenToOutputMomentum = new double[nHidden][nOutput];
        HiddenMomentum = new double[nHidden];
        OutputMomentum = new double[nOutput];
        // error terms in hidden and output layers
        HiddenErrors = new double[nHidden];
       	Errors = new double[nOutput];
        // bias in hidden and output layers
        HiddenThresholds = new double[nHidden];
        OutputThresholds = new double[nOutput];
        // derivative values of activation functions in
        // hidden and output layers
        HiddenDeriv = new double[nHidden];
        OutputDeriv = new double[nOutput];
        // target output
        DesiredOutput = new double[nOutput];
        Eta = World.TASK_PARA_ARRAY[ World.NET_LRATE ];
        Momentum = World.TASK_PARA_ARRAY[ World.NET_MOMENTUM ];
        // initialize weights.
        initWeights( 0 );
    }

    /** new constructor.
     *  @param inputNum number of input units.
     *  @param netIdx net index.
     *  @param global the global settings.
     *  @param netType net type. */
    public BPNet( int inputNum, int netIdx, Global global, int netType )
    {
        nInput = inputNum;
        this.global = global;
        if( netType == Global.ACS ){
            //nInput = global.getIdnInputNum( netIdx );
            nHidden = global.getIdnHiddenNum( netIdx );
            //nOutput = global.getIdnOutputNum( netIdx );
            nOutput = global.getAcsEntirePhysicOutputNum();
            Eta = global.getIdnBPLearningRate( netIdx );
            Momentum = global.getIdnBPMomentum( netIdx );
            inputDVs = global.getIdnInputDimVals( netIdx );
            outputDVs = global.getIdnOutputDimVals( netIdx );
            if( global.getAcsCrossCombineMode(netIdx, 1 ) == Global.VAR_COMB_MODE ){
                SUCC_C3 = global.getAcsCrossCombineVarParaAbove( netIdx )[Global.BL];
                SUCC_C4 = global.getAcsCrossCombineVarParaAbove( netIdx )[Global.BL];
            }
            if( global.isRTOn() ){
                BLPT = global.getPT_BL();
                BLAT = global.getAT_BL();
                BLDT = global.getDT_BL();
            }

            //System.out.println( "# of BL output = " + nOutput );
            //System.exit( 1 );

        }
        else{  // AMN of NACS
            //nInput = global.getAmnInputNum( netIdx );
            nHidden = global.getAmnHiddenNum( netIdx );
            nOutput = global.getAmnOutputNum( netIdx );
            Eta = global.getAmnBPLearningRate( netIdx );
            Momentum = global.getAmnBPMomentum( netIdx );
            inputDVs = global.getAmnInputDimVals( netIdx );
            outputDVs = global.getAmnOutputDimVals( netIdx );
            if( global.isRTOn() )   // need be updated.
            {
                BLPT = global.getPT_BL();
                BLAT = global.getAT_BL();
                BLDT = global.getDT_BL();
            }
        }
        // start points of each dim in 1-dim state.
        outputOffsets = new short[ outputDVs.length ];
        //outputOffsets[ 0 ] = 0;
        //for( int i = 1; i < outputDVs.length; i++ )
             //outputOffsets[i] = (short)(outputOffsets[i-1] + outputDVs[i-1]);
        CommonMethods.calcOffsets( outputOffsets, outputDVs );
        // weight vectors
        InputToHiddenWeights = new double[nInput][nHidden];
        HiddenToOutputWeights = new double[nHidden][nOutput];
        ItoHWeights = new double[nInput][nHidden];
        HtoOWeights = new double[nHidden][nOutput];
        // momentum vectors
        InputToHiddenMomentum = new double[nInput][nHidden];
        HiddenToOutputMomentum = new double[nHidden][nOutput];
        HiddenMomentum = new double[nHidden];
        OutputMomentum = new double[nOutput];
        // value of input, hidden and output vectors
        Input = new double[nInput];
        Hidden = new double[nHidden];
        Output = new double[nOutput];
        // bias in hidden and output layers
        HiddenThresholds = new double[nHidden];
        OutputThresholds = new double[nOutput];
        // derivative values of activation functions in
        // hidden and output layers
        HiddenDeriv = new double[nHidden];
        OutputDeriv = new double[nOutput];
        // error terms in hidden and output layers
        HiddenErrors = new double[nHidden];
       	Errors = new double[nOutput];
        // desired output
	       DesiredOutput = new double[nOutput];
        // initialize weights.
       	initWeights( 0 );
    }

    /** reinitializes the weights. */
    public void reinitWeights()
    {
       initWeights( 0 );
    }

    /** restores the weights. */
    public void restoreInitWeights()
    {
       initWeights( 1 );
    }

/***************************** GET AND SET METHODS ****************************/
    /**
     * returns the nubmer of input units.
     * @return the nubmer of input units. */
    public int getnInput()
    {
       return nInput;
    }

    /**
     * returns the nubmer of hidden units.
     * @return the nubmer of hidden units. */
    public int getnHidden()
    {
       return nHidden;
    }

    /**
     * returns the nubmer of output units.
     * @return the nubmer of output units. */
    public int getnOutput()
    {
      	return nOutput;
    }

    /**
     * sets the learning rate.
     * @param value the value to set. */
    public void setLearningRate( double value )
    {
        Eta = value;
    }

    /**
     * sets the momentum.
     * @param value the value to set. */
    public void setMomentum( double value )
    {
        Momentum = value;
    }

    /**
     * returns current input.
     * @param fillMe the array to store current input. */
    public void getInput( double[] fillMe )
    {
        for( int i = 0; i < nInput; i++ )
             fillMe[i] = Input[i];
    }

    /**
     * returns current output.
     * @return current output in the format of an array. */
    public double[] getOutput()
    {
        return Output;
    }

    /**
     * returns current output.
     * @param out the array to store current output. */
    public void getOutput( double[] out )
    {
       	for( int i = 0; i < nOutput; i++ )
             out[ i ] = Output[i];
    }

    /**
     * returns a particular output.
     * @param idx the index in the output array.
     * @return the particular output. */
    public double getOutput( int idx )
    {
       return Output[ idx ];
    }

    /**
     * returns a particular output.
     * @param dimIdx the index on a dimension in the output.
     * @param offset the offset inside the dimension.
     * @return the particular output. */
    public double getOutput( int dimIdx, int offset )
    {
       //if( outputOffsets[ dimIdx ] + offset >= Output.length ||
       //    outputOffsets[ dimIdx ] + offset < 0 )
           //System.out.println( "dim offset = " + outputOffsets[ dimIdx ] + ", offset = " + offset +
           //                    ", output len = " + Output.length );

       return Output[ outputOffsets[ dimIdx ] + offset ];
    }

    /**
     * Returns a part of current output.
     * @param out the array to store current output. */
    public void getOutput( double[] fillMe, int start, int length )
    {
       System.arraycopy( Output, start, fillMe, 0, length );
    }

    /**
     * Sets current input.
     */
    public void setInput( int[] newInputs )
    {
       	for( int i = 0; i < nInput; i++ )
             Input[i] = newInputs[i];
    }

    /**
     * set current input.
     */
    public void setInput( double[] newInputs )
    {
       	for( int i = 0; i < nInput; i++ )
             Input[i] = newInputs[i];
    }

    /**
     * set particular element in the input.
     */
    public void setInput( int index, double value )
    {
       	Input[index] = value;
    }

    /**
     * set the desired output.
     */
    public void setDesiredOutput( double[] dOutputs )
    {
       	for( int i = 0; i < nOutput; i++ )
             DesiredOutput[i] = dOutputs[i];
    }

    /**
     * set one desired element in the output.
     */
    public void setDesiredOutput( int index, double value )
    {
       	DesiredOutput[index] = value;
    }

    /**
     * set the desired output with an array of desired acions.
     */
    public void setDesiredOutput( short[] action )
    {
        for( int i = 0; i < outputDVs.length; i++ )
          for( int j = 0; j < outputDVs[i]; j++ )
               DesiredOutput[ outputOffsets[i] + j ] = (j==action[i]) ? 1 : 0;
    }

    /**
     * set the desired output with an array of desired features.
     */
    public void setDesiredOutput( Feature[] output )
    {
        for( int i = 0; i < outputDVs.length; i++ )
          for( int j = 0; j < outputDVs[i]; j++ )
               DesiredOutput[ outputOffsets[i] + j ] = output[i].getActivations()[j];
    }

    public double getSumSqErrors()
    {
       	return sumSqErrors;
    }

    /**
     * get the number of positive match.
     */
    public double getPM()
    {
        return PM;
    }

    /**
     * get the number of negative match.
     */
    public double getNM()
    {
        return NM;
    }

    /**
     * get the success rate.
     */
    public double getSuccRate()
    {
       succRate = ( SUCC_C3 + PM ) / ( SUCC_C4 + PM + NM );
       return succRate;
    }

    /**
     * get the response time.
     */
    public int getResponseTime()
    {
       return BLRT;
    }

    /**
     * resets the matches.
     */
    public void resetMatches()
    {
        PM = 0.0;
        NM = 0.0;
    }

    /**
     * updates the matches.
     */
    public void updateMatches( boolean flag )
    {
        if( flag )
            PM += 1.0;
        else
            NM += 1.0;
    }

    /**
     * calculate the actual output given current input.
     */
    public void forwardPass()
    {
       	computeHiddenActivation();
        computeOutputActivation();
    }

    /**
     * update all of the weights inside the network triggered by the difference
     * between desired output and the actual output.
     */
    public void backwardPass()
    {
       	computeErrors();
        computeHiddenErrors();
        modifyHiddenToOutput();
        modifyInputToHidden();
    }

    /**
     * calculate the response time.
     */
    public void calcRT( int BLAT )
    {
        this.BLAT = BLAT;
        BLRT = global.getTaskClarion().getBLOverallRT( BLPT, BLDT, BLAT );
    }

/************************** PROTECTED METHODS *********************************/
    /********************* forward pass methods **************************/
    /**
     * calculates the hidden derivative.
     */
    protected void computeHiddenActivation()
    {
       	double sum,v;
       	for( int j = 0; j < nHidden; j++ ){
             sum = HiddenThresholds[j];
             for( int i = 0; i < nInput; i++ )
	          sum += Input [i] * InputToHiddenWeights[i][j];
             v = activation( sum );
             Hidden[j] = v;
             HiddenDeriv[j] = derivative(v);
        }
    }

    /**
     * calculates the output derivative.
     */
    protected void computeOutputActivation ()
    {
       	double sum,v;
       	for( int j = 0; j < nOutput; j++ ){
             sum = OutputThresholds[j];
             for( int i = 0; i < nHidden; i++ )
                  sum += Hidden[i] * HiddenToOutputWeights[i][j];
             v = activation(sum);
             Output[j] = v;
             OutputDeriv[j] = derivative(v);
       	}
    }

    /****************** backward pass methods *****************************/
    /**
     * computes the errors of output layer.
     */
    protected void computeErrors()
    {
        sumSqErrors = 0;
       	for( int i = 0; i < nOutput; i++ ){
             double err = DesiredOutput[i] - Output[i];
             Errors[i] = err * OutputDeriv[i];
             sumSqErrors += err * err;
        }
       	sumSqErrors = sumSqErrors / 2;
        if( debug )
            System.out.println( sumSqErrors );
    }

    /**
     * computes the errors of hidden layer.
     */
    protected void computeHiddenErrors ()
    {
       	double h_err;
       	for( int i = 0; i < nHidden; i++ ){
             h_err = 0.0;
             for( int j = 0; j < nOutput; j++ )
     	          h_err += HiddenToOutputWeights[i][j] * Errors[j];
             HiddenErrors[i] = h_err * HiddenDeriv[i];
        }
    }

    /**
     * modifies the weights of hidden to output layer
     */
    protected void modifyHiddenToOutput ()
    {
        double x;
        for( int j = 0; j < nOutput; j++ ) {
       	     if( Math.abs( Errors[j] ) > RZero ) {
              	 x = Errors[j];
                 // udpate HiddenToOutputWeights
                 for( int i = 0; i < nHidden; i++ ){
                      if( Momentum > RZero ){
                            /* momentum */
                          HiddenToOutputMomentum[i][j] = Eta * x * Hidden[i] +
                                                         Momentum * HiddenToOutputMomentum[i][j];
                          HiddenToOutputWeights[i][j] += HiddenToOutputMomentum[i][j];
                      }
                      else
                          HiddenToOutputWeights[i][j] += Eta * x * Hidden[i];
                 }
                 // update OutputThresholds
                 if( Momentum > RZero ){
                     OutputMomentum[j] =  Eta * x + Momentum * OutputMomentum[j];
                     OutputThresholds[j] += OutputMomentum[j];
                 }
                 else
                     OutputThresholds[j] += Eta * x;
      	    }
       	}
    }

    /**
     * modifies the weights of input to hidden layer
     */
    protected void modifyInputToHidden ()
    {
       	double x;
       	for( int j = 0; j < nHidden; j++ ){
             if( Math.abs( HiddenErrors[j] ) > RZero ){
                 x = HiddenErrors[j];
                 // update the InputToHiddenWeights
            	 for(int i = 0; i < nInput; i++){
                     if( Momentum > RZero ){
                     	 InputToHiddenMomentum[i][j] = Eta * x* Input[i]
                                                       + Momentum * InputToHiddenMomentum[i][j];
      			 InputToHiddenWeights[i][j] += InputToHiddenMomentum[i][j];
                     }
                     else
               		 InputToHiddenWeights[i][j] += Eta * x * Input[i];
               	 }
                 // upate the HiddenThresholds
       		 if( Momentum > RZero){
               	     HiddenMomentum[j] = Eta * x + Momentum * HiddenMomentum[j];
               	     HiddenThresholds[j] += HiddenMomentum[j];
                 }
                 else
        	     HiddenThresholds[j] += Eta * x;
             }
     	}
    }

/****************************** PRIVATE METHODS *******************************/
    private void initWeights( int mode )
    {
        for(int i = 0; i < nInput; i++)
           for(int j = 0; j < nHidden; j++)
               InputToHiddenMomentum[i][j] = Momentum;

      	for( int i = 0; i < nHidden; i++){
	     HiddenThresholds[i] = randomThresholds();
             //HiddenThresholds[i] = 1;
             HiddenMomentum[i] = 0;
      	}

        for( int i = 0; i < nHidden; i++)
            for(int j = 0; j < nOutput; j++)
                HiddenToOutputMomentum[i][j] = Momentum;

      	for( int i = 0; i < nOutput; i++ ){
             OutputThresholds[i] = randomThresholds();
             //OutputThresholds[i] = 1;
             OutputMomentum[i] = 0;
       	}

        if( mode == 0 ){
       	    for( int i = 0; i < nInput; i++ )
             	for( int j = 0; j < nHidden; j++ ){
                   InputToHiddenWeights[i][j] = randomWeights();
                   ItoHWeights[i][j] = InputToHiddenWeights[i][j];
               }
            for (int i = 0; i < nHidden; i++)
                for (int j = 0; j < nOutput; j++){
              		     HiddenToOutputWeights[i][j] = randomWeights();
                     HtoOWeights[i][j] = HiddenToOutputWeights[i][j];
                }
        }
        else{
       	    for( int i = 0; i < nInput; i++ )
            	for( int j = 0; j < nHidden; j++ )
      	   	       InputToHiddenWeights[i][j] = ItoHWeights[i][j];
            for(int i = 0; i < nHidden; i++)
              for(int j = 0; j < nOutput; j++)
                  HiddenToOutputWeights[i][j] = HtoOWeights[i][j];
        }
    }

    private double randomWeights()
    {
       	double sign,value;
        if ( Math.random() < 0.5 )
             sign = LINITWEIGHT;
        else
             sign = HINITWEIGHT;
        value = Math.random();
        return( value * sign );
    }

    private double randomThresholds()
    {
       	double sign,value;
        if ( Math.random() < 0.5 )
             sign = LINITTHRESHOLD;
        else
             sign = HINITTHRESHOLD;
        value = Math.random();
        return( value * sign );
    }

    // activation function
    private double activation(double value)
    {
       	// can switch other activation functions here
       	// simple sigmoid function
        //System.out.println("the combined value = " + value );

       	return (1.0 / (1.0 + Math.exp( -value)));
    }

    // derivative of the activation function
    private double derivative( double value )
    {
       	return( value * ( 1 - value ) );
    }
}
