package clarion.common;
import java.util.Arrays;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to impliment the concept of <em>dimension</em> in CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public Feature( String label, short type, short order, short[] con, Global g ):
 * constructor.
   <li> public Feature( String label, short type, short order, double[] actVals, Global g ):
   constructor.
   <li> public Feature( String label, short type, short order, short length, short[] vals, Global g ):
   constructor.
 * <li> public boolean isContained( Feature other ): is to check if this feature
 * is contained by other feature.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is an atomic class in current system for knowledge representation.
 * <li> Usage of this class. <br>
 * It is instantiated either inside a concrete chunk object or anywhere needs
 * construct a <em>Feature</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class Feature implements Cloneable {

   /** gloabl settings. */
   protected Global  global;
   /** the label of the feature. */
   protected String   label;
   /** the type of the feature. */
   protected short    type;
   /** order in the whole feature list. */
   protected short    order;
   /** the length of this feature. */
   protected short    length;
   /** the active values this feature has.   */
   protected short[]  values;
   /** activation of each value in feature. */
   protected double[] activations;
   /** used for WM item. */
   protected boolean  flag = false;

   /** constructor used for WM flags. */
   public Feature( String label, int length, int value )
   {
      this.label = label;
      this.length = (short)length;
      values = new short[1];
      values[0] = (short)value;
      flag = true;
   }

   /** constructor.
    *  @param label the feature label.
    *  @param type the feature type.
    *  @param order the order in the chunk this feature is in.
    *  @param con the content of the feature in terms of an array where each
    *  element indicates a dimensional value is active or not.
    *  @g the associated global settings. */
   public Feature( String label, short type, short order, short[] con, Global g )
   {
      this.label = label;
      this.type = type;
      this.order = order;
      length = ( con == null ) ? 0 : (short)con.length;
      int valNum = 0;
      short[] content = new short[ length ];
      activations = new double[ length ];
      for( int i = 0; i < length; i++ )
           if( con[ i ] == 1 ){
               content[ valNum++ ] = (short)i;
               activations[ i ] = 1.0;
           }
      values = new short[ valNum ];
      for( int i = 0; i < valNum; i++ )
           values[ i ] = content[ i ];
      Arrays.sort( values );
      global = g;
   }

   /** constructor.
    *  @param label the feature label.
    *  @param type the feature type.
    *  @param order the order in the chunk this feature is in.
    *  @param actVals the content of the feature in terms of an array of activations
    *  each value in the dimension has.
    *  @g the associated global settings.  */
   public Feature( String label, short type, short order, double[] actVals, Global g )
   {
      this.label = label;
      this.type = type;
      this.order = order;
      length = (short)actVals.length;
      int valNum = 0;
      short[] content = new short[ length ];
      activations = new double[ length ];
      for( int i = 0; i < length; i++ ){
           activations[ i ] = actVals[ i ];
           if( activations[ i ] > 0.0 )
               content[ valNum++ ] = (short)i;
      }
      values = new short[ valNum ];
      for( int i = 0; i < valNum; i++ )
           values[ i ] = content[ i ];
      Arrays.sort( values );
      global = g;
   }

   /** constructor.
    *  @param label the feature label.
    *  @param type the feature type.
    *  @param order the order in the chunk this feature is in.
    *  @param length the length of the dimension.
    *  @param vals the content of the feature in terms of an array of active
    *  values this feature has.
    *  @g the associated global settings.   */
   public Feature( String label, short type, short order, short length, short[] vals, Global g )
   {
      this.label = label;
      this.type = type;
      this.order = order;
      this.length = length;
      activations = new double[ length ];
      values = ( vals != null ) ? new short[ vals.length ] : new short[0];
      if( vals != null ){
          for( int i = 0; i < vals.length; i++ ){
               values[ i ] = vals[ i ];
               activations[ values[i] ] = 1.0;
          }
          Arrays.sort( values );
      }
      global = g;
   }

   /**
    * Compares two features is the same or not.
    */
   public boolean isTheSameFeature( Feature other )
   {
      return getType()==other.getType() && getOrder()==other.getOrder();
   }

   /**
    * returns the feature label.
    * @return the feature label. */
   public String getLabel()
   {
      return label;
   }

   /**
    * returns the feature type.
    * @return the feature type. */
   public int getType()
   {
      return type;
   }

   /**
    * returns the feature order.
    * @return the feature order. */
   public int getOrder()
   {
      return order;
   }

   /**
    * returns the feature length.
    * @return the feature length. */
   public int getLength()
   {
      return length;
   }

   /**
    * returns the number of active values.
    * @return the number of active values. */
   public int getValueNum()
   {
      return values.length;
   }

   /**
    * returns active values.
    * @return active values. */
   public short[] getValues()
   {
      return values;
   }

   /**
    * returns the active values.
    * @param vals the array to store the active values. */
   public void getValues( short[] vals )
   {
      for( int i = 0; i < vals.length; i++ )
           vals[ i ] = values[ i ];
   }

   /**
    * returns the content in terms of an array where each element indicates the
    * corresponding dimensional value is active or not.
    * @return the content of this feature. */
   public short[] getContent()
   {
      short[] content = new short[ length ];
      for( int i = 0; i < length; i++ )
           content[ i ] = 0;
      for( int i = 0; i < values.length; i++ )
           content[ values[ i ] ] = 1;
      return content;
   }

   /**
    * return the content in terms of an array where each element indicates the
    * corresponding dimensional value is active or not.
    * @param content the array to store the content of this feature. */
   public void getContent( short[] content )
   {
      for( int i = 0; i < length; i++ )
           content[ i ] = 0;
      for( int i = 0; i < values.length; i++ )
           content[ values[ i ] ] = 1;
   }

   /**
    * returns the activations.
    * @return the activations. */
   public double[] getActivations()
   {
      return activations;
   }

   /**
    * returns the activations.
    * @return vals the array to store the activations. */
   public void getActivations( double[] vals )
   {
      for( int i = 0; i < vals.length; i++ )
           activations[ i ] = vals[ i ];
   }

   /**
    * returns the activations.
    * @return vals the array to store the activations. */
   public void setValues( short[] arr )
   {
      values = CommonMethods.copy( arr );
      Arrays.sort( values );
   }

   /** returns if this feature is contained in other feature. */
   public boolean isContained( Feature other )
   {
       if( values.length > other.getValueNum() )
           return false;
       short[] con1 = this.getContent();
       short[] con2 = other.getContent();
       for( int i = 0; i < length; i++ )
            if( con1[i] == 1 && con2[i] == 0 )
                return false;
       return true;
   }

/************************* OVERRIDE METHODS ************************************/
   /** methods from class <em>Object</em> */
   public String toString()
   {
      String[] names = global.getFeatureValueNames( type, order );
      //String str = "( " + label + ":";
      String str = "( ";
      for( int i = 0; i < values.length; i++ )
           str += names[ values[i] ] + " ";
      str += ")";
      return str;
   }

   /** methods from class <em>Object</em> */
   public int hashCode()
   {
      /*
      int sum = 0;
      for( int i = 0; i < values.length; i++ )
           sum += values[ i ];
      return sum;
      */
      Arrays.sort( values );
      int num = Math.min( 32, length );
      int code = 0;
      for( int i = 0; i < num; i++ )
           code = (code <<= 1) + ((Arrays.binarySearch(values, (short)i) >= 0) ? 1 : 0);
      return code;
   }

   /** methods from class <em>Object</em> */
   public boolean equals( Object other )
   {
      if( other == null || !(other instanceof Feature ) )
          return false;
      Feature otherFeat = (Feature)other;
      if( label.compareTo( otherFeat.getLabel()) != 0 )
          return false;
      if( type != otherFeat.getType() )
          return false;
      if( order != otherFeat.getOrder() )
          return false;
      if( values.length != otherFeat.getValueNum() )
          return false;
      Arrays.sort( otherFeat.getValues() );
      return Arrays.equals( values, otherFeat.getValues());
   }

   /** methods from class <em>Object</em> */
   public Object clone()
   {
      Feature feature = flag ? new Feature( label, length, values[0] ) :
                               new Feature( label, type, order, length, values, global );
      return feature;
   }
}