package clarion.coordinate;
import clarion.system.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the generic method of coordinating actions of EXTERNAL
 * and goal structure.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void coordinate(): coordination of EX and GS as described in the
 * description.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * In CLARION model, there are two types of coordination: one is coordination
 * between different external actions, the other is coordination between external
 * actions and goal action. The class implements the latter.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>ACS</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class ExGScoordinator {
   final int DO_NOTHING = -2, NO_AVAILABLE = -1;
   /** global settings. */
   protected Global global;
   /** the task agent. */
   protected TaskAgent agent;
   /** the performed action type. */
   protected short performActionType;
   /** the chosen action. */
   protected short chosenActions[][][][];
   /** the probability of using external action. */
   protected double exProb;
   /** the probability of using goal action. */
   protected double gsProb;

   // constructor for stochastic combination method.
   // mode indicates fixed or varied. = 0: fixed; = 1: varied.
   public ExGScoordinator(){}

   /** constructor.
    *  @param agent the task agent.
    *  @param global the global settings. */
   public ExGScoordinator( TaskAgent agent, Global global )
   {
      this.agent = agent;
      this.global = global;
      exProb = global.getExternCorConstant();
      gsProb = global.getGoalCorConstant();
   }

   /** Sets the chosen action.
    *  @param actions the chosen action to set. */
   public void setChosenActions( short[][][][] actions )
   {
      chosenActions = actions;
   }

   /** coordination of EX and GS as described in the description. */
   public void coordinate()
   {
      int exAction = checkExActions( chosenActions );
      int gsAction = NO_AVAILABLE;
      int wmAction = NO_AVAILABLE;
      if( global.getGsOption() != Global.GS_NON )
          gsAction = checkGsAction( chosenActions );
      if( global.isWmOn() )
          wmAction = checkWmAction( chosenActions );
      // decide type of action being performed.
      performActionType = Global.EMPTY;
      if( exAction == NO_AVAILABLE || exAction == DO_NOTHING ){
          if( gsAction >= 0 )
              performActionType = Global.GS_TYPE;
      }
      if( gsAction == NO_AVAILABLE || gsAction == DO_NOTHING ){
          if( exAction >= 0 )
              performActionType = Global.EX_TYPE;
      }
      if( ( exAction == NO_AVAILABLE || exAction == DO_NOTHING ) &&
          ( gsAction == NO_AVAILABLE || gsAction == DO_NOTHING ) ){
          if( wmAction >= 0 )
              global.getACS().performWmAction();
          return;
      }

      //System.out.println( "before, action type = " + performActionType );

      if( performActionType == Global.EMPTY )
          randomCoordinate();

      //System.out.println( "after, action type = " + performActionType );

      if( performActionType == Global.EX_TYPE ){
          for( int i = 0; i < chosenActions.length; i++ ){
               if( global.isExNet(i) )
                   agent.performAction( i, chosenActions[i] );
          }
          if( wmAction >= 0 )
              global.getACS().performWmAction();
          if( gsAction >= 0 )
              global.getACS().performGsAction();
      }
      if( performActionType == Global.GS_TYPE ){
          if( wmAction >= 0 )
              global.getACS().performWmAction();
          if( gsAction >= 0 )
              global.getACS().performGsAction();
      }
   }

   /** using "random selection" method to select EX or GS as currently chosen action
    *  type provided that both chosen actions from EX and GS are available and not
    *  do nothing. */
   private void randomCoordinate()
   {
      double randVal = Math.random();
      if( randVal < exProb )
          performActionType = Global.EX_TYPE;
      else
          performActionType = Global.GS_TYPE;
   }

   /** check if EX action is DO_NOTHING or NOT_AVAILABLE. */
   private int checkExActions( short[][][][] actions )
   {
      int value = NO_AVAILABLE;
      for( int i = 0; i < actions.length; i++ )
           if( global.isExNet(i) ){
               for( int j = 0; j < actions[i].length; j++ )
                    if( actions[i][j] != null )
                        for( int k = 0; k < actions[i][j].length; k++ )
                             if( actions[i][j][k] != null && actions[i][j][k].length > 0 ) //not do-nothing.
                                 value = 1;
           }
      return value;
   }

   /** check if GS action is DO_NOTHING or NOT_AVAILABLE. */
   private int checkGsAction( short[][][][] actions )
   {
      int value = NO_AVAILABLE;
      for( int i = 0; i < actions.length; i++ )
           if( global.isGsNet(i) ){
               if( actions[i][Global.ACTION_NORMAL][0] != null &&
                   actions[i][Global.ACTION_NORMAL][0][0] != 0 )   // not do-nothing.
                   value = 1;
           }
      return value;
   }

   /** check if WM action is DO_NOTHING or NOT_AVAILABLE. */
   private int checkWmAction( short[][][][] actions )
   {
      int value = NO_AVAILABLE;
      for( int i = 0; i < actions.length; i++ )
           if( global.isWmNet(i) ){
               if( actions[i][Global.ACTION_NORMAL][0] != null &&
                   actions[i][Global.ACTION_NORMAL][0][0] != 0 )   // not do-nothing.
                   value = 1;
           }
      return value;
   }
}