package clarion.coordinate;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;
import clarion.acs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the weight-sum combination of outputs from both
 * levels.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public short[][] combination(): Combines the outputs from both levels.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * The class implements the weight-sum combination of outputs from both levels,
 * one of the three basic combination methods.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>AcsNet</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class WeightSumComb{

   /** the number of components. */
   int componentNum;
   /** the number of dimensions. */
   int dimNum;
   /** weights from different components in a network. */
   double weightArr[];
   /** array of beta values. */
   double betaArr[];
   /** array of success rates. */
   double succArr[];
   /** An array of VCGSet to select active values in each dimension. */
   VCGSet vcgSets[];
   /** arrays of VCGSet to store the active values in each dimension from each
    *  net component such as IDN, RER, IRL and FR. */
   VCGSet compVcgSets[][];

   /** constructor, for old version.
    *  @param mode = 0: fixed method which means the weights are fixed,
              mode = 1: variable method which means the weights are got by calculating.
       @param paramArr the array of parameter values.
    *         mode = 0: carries the values of weights.
    *         mode = 1: carries the values of beta.
    *  @param temperature the temperature for action decision. */
   public WeightSumComb( int mode, double[] paramArr, double temperature )
   {
      componentNum = paramArr.length;
      // initialize the parameters.
      if( mode == Global.FIX_COMB_MODE )  // fixed mode
      {
          weightArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               weightArr[ i ] = paramArr[ i ];
      }
      else               // varied mode
      {
          weightArr = new double[ componentNum ];
          betaArr = new double[ componentNum ];
          succArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               betaArr[ i ] = paramArr[ i ];
      }
   }

   /** Returns the weights.
    *  @param arr the array to store the weights. */
   public void getWeights( double[] arr )
   {
      for( int j = 0; j < componentNum; j++ )
           arr[ j ] =  weightArr[ j ];
   }

   /** Sets the success rates. used in varied mode.
    *  @param arr the array to set the success rates. */
   public void setSuccRate( double[] arr )
   {
       for( int j = 0; j < componentNum; j++ )
            succArr[ j ] = arr[ j ];
   }

   public void setVCGSets( VCGSet[][] compVcgSets )
   {
      this.compVcgSets = compVcgSets;
      // initialize the VCGSets.
      dimNum = compVcgSets[0].length;
      vcgSets = new VCGSet[ dimNum ];
      for( int i = 0; i < dimNum; i++ ){
           int type = compVcgSets[0][i].getType();
           int vcgNum = compVcgSets[0][i].getVCGNum();
           short[] elemNums = compVcgSets[0][i].getElemNums();
           vcgSets[i] = new VCGSet( type, vcgNum, elemNums );
      }
   }

   /** Calculates the weights. */
   public void calcWeights()
   {
       double sum = 0;
       for( int j = 0; j < componentNum; j++ )
             sum += betaArr[ j ] * succArr[ j ];
       for( int j = 0; j < componentNum; j++ )
            weightArr[ j ] = betaArr[ j ] * succArr[ j ] / sum;
   }

   /** Combines the outputs from both levels. this is the new method since the
    *  new format of action is two-dimensional in stead of one-dimensional.
    *  @return the array to store the combined action. */
   public short[][] combination()
   {
      short[][] chosenAction = new short[ dimNum ][];
      for( int i = 0; i < dimNum; i++ ){
           int length = vcgSets[i].getFullLength();
           double[] activations = new double[ length ];
           double[][] compActivations = new double[ componentNum ][ length ];
           for( int j = 0; j < componentNum; j++ )
                compActivations[j] = vcgSets[i].getActivations();
           for( int j = 0; j < length; j++ )
            for( int k = 0; k < componentNum; k++ ){
                  double val = weightArr[k] * compActivations[k][j];
                  activations[ j ] += val;
            }
           vcgSets[i].setActivations( activations );
           chosenAction[i] = vcgSets[i].decideActiveValues();
      }
      return chosenAction;
   }
}