package clarion.ms;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.acs.gs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the MS subsystem of CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void offTrainDriveNet( int netIdx ): Offline training a specific
 * drive network.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements MS subsystem, the motivational subsystem, and on the top
 * of MS implementation structure and all the other classes relevant to MS are
 * declared inside it.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>Clarion</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class MS {

   /** number of drive networks: low-level primary drives, high-level primary
    *  dirves and secondary drives. */
   protected final int DR_NET_NUM = 3;
   /** global settings. */
   protected Global global;
   /** goal structure, the one in ACS.  */
   protected GoalStructure gs;
   /** goal actions for ACS. */
   protected int acsGoalAction;
   /** goal actions for MCS. */
   protected int mcsGoalAction;
   /** the goals for ACS and MCS by performing the corresponding action. */
   protected GoalChunk acsGoal, mcsGoal;
   /** the sensory input to the drive networks. */
   protected double[] sensoryInput;
   /** the drive outputs from the drive networks. */
   protected double[] driveOutput[] = new double[DR_NET_NUM][];
   /** the drive BP networks. */
   protected BPNet[] driveNets = new BPNet[DR_NET_NUM];

   /** constructor.
    *  @param global the assoiciated global settings.
    *  @param gs the used goal structure. */
   public MS( Global global, GoalStructure gs )
   {
      this.global = global;
      this.gs = gs;
      global.setMS( this );
   }

   /** Sets ACS goal action.
    *  @param action the action to set. */
   public void setAcsGoalAction( int action )
   {
      acsGoalAction = action;
   }

   /** Sets MCS goal action.
    *  @param action the action to set. */
   public void setMcsGoalAction( int action )
   {
      mcsGoalAction = action;
   }

   /** Sets sensory input.
    *  @param input the senosry input to set. */
   public void setSensoryInput( double[] input )
   {
      sensoryInput = input;
   }

   /** Returns the ACS goal in terms of goal chunk
    *  @return the ACS goal in terms of goal chunk. */
   public GoalChunk getAcsGoal()
   {
      return acsGoal;
   }

   /** Returns the MCS goal in terms of goal chunk
    *  @return the MCS goal in terms of goal chunk. */
   public GoalChunk getMcsGoal()
   {
      return mcsGoal;
   }

   /** Returns the drive strengths from the specific network.
    *  @param netIdx the network index: low-level, high-level and secondary. */
   public double[] getDriveStrength( int netIdx )
   {
      driveNets[netIdx].setInput( sensoryInput );
      driveNets[netIdx].forwardPass();
      return driveNets[netIdx].getOutput();
   }

   /** Offline training a specific drive network.
    *  @param netIdx the network index: low-level, high-level and secondary. */
   public void offTrainDriveNet( int netIdx )
   {
      driveNets[netIdx].forwardPass();
      driveNets[netIdx].backwardPass();
   }
}