package clarion.nacs;
import clarion.common.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the probabilistic backpropagation learning algorithm
 * used in AFN and SFN.
 *
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void forward(): Computes the actual outputs.
 * <li> public void backward(): Updates the weights.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the AFN and SFN in abstract episodic memory in NACS subsystem.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>AbsEM</em> object.
 * </ul>
 * @author Xi Zhang
 */

public class ProbBPNet {

   /** learning rate. */
   protected double alpha;
   /** the number of input units. */
   protected int INPUT_NUM;
   /** the number of output units. */
   protected int OUTPUT_NUM;
   /** the inputs */
   protected double inputs[];
   /** the outputs. */
   protected double outputs[];
   /** the desired outputs. */
   protected double desiredOutputs[];
   /** the weights. */
   protected double weights[][];

   /** constructor.
    *  @param num1 the number of inputs.
    *  @param num2 the number of outputs.
    *  @param lr learning rate. */
    public ProbBPNet( int num1, int num2, double lr )
   {
      INPUT_NUM = num1;
      OUTPUT_NUM = num2;
      alpha = lr;
      inputs = new double[ INPUT_NUM ];
      outputs = new double[ OUTPUT_NUM ];
      desiredOutputs = new double[ OUTPUT_NUM ];
      weights = new double[ INPUT_NUM ][ OUTPUT_NUM ];
      initWeights();
   }

  /** Returns a specific output.
   *  @param idx the index to the specific output.
   *  @return the specific output. */
   public double getOutput( int idx )
   {
      return outputs[ idx ];
   }

  /** Returns the outputs.
   *  @param arr the array to store the outputs.
   *  @return the outputs. */
   public void getOutputs( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           arr[ i ] = outputs[ i ];
   }

  /** Sets the inputs.
   *  @param arr the array used to set the inputs. */
   public void setInputs( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           inputs[ i ] = arr[ i ];
   }

  /** Sets the inputs in terms of an array of activations.
   *  @param arr the array used to set the inputs. */
   public void setInputs( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           inputs[ i ] = arr[ i ];
   }

  /** Sets the desired outputs.
   *  @param arr the array used to set the desired outputs. */
   public void setDesiredOutputs( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           desiredOutputs[ i ] = arr[ i ];
   }

  /** Sets the desired outputs  in terms of an array of activations.
   *  @param arr the array used to set the desired outputs. */
   public void setDesiredOutputs( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           desiredOutputs[ i ] = arr[ i ];
   }

   /** Computes the actual outputs. */
   public void forward()
   {
      for( int i = 0; i < OUTPUT_NUM; i++ )
           outputs[ i ] = 0.0;
      for( int i = 0; i < INPUT_NUM; i++ )
        for( int j = 0; j < OUTPUT_NUM; j++ )
             outputs[ j ] += inputs[ i ] * weights[ i ][ j ];
      for( int i = 0; i < OUTPUT_NUM; i++ )
           outputs[ i ] = 1.0 / ( 1.0 + java.lang.Math.exp( -outputs[ i ] ) );
   }

   /** Updates the weights. */
   public void backward()
   {
      //System.out.println( "stateIdx = " + stateIdx + ", actionIdx = " + actionIdx
      //                    + ", weight before = " + weights[stateIdx][actionIdx] );
      for( int i = 0; i < INPUT_NUM; i++ )
        for( int j = 0; j < OUTPUT_NUM; j++ )
        {
             double deltaW = alpha * ( desiredOutputs[ j ] - outputs[ j ] ) * inputs[ i ];
             weights[ i ][ j ] += deltaW;
        }
      //System.out.println( "weight after = " + weights[stateIdx][actionIdx] );
   }

   /** initialization routine. */
   public void initWeights()
   {
      double HINITWEIGHT = 0.01;
      double LINITWEIGHT = -0.01;
      for( int i = 0; i < INPUT_NUM; i++ )
        for( int j = 0; j < OUTPUT_NUM; j++ )
        {
             double sign, value;
             if ( java.lang.Math.random() < 0.5 )
                  sign = LINITWEIGHT;
             else
                  sign = HINITWEIGHT;
             value = java.lang.Math.random();
             weights[ i ][ j ] =  value * sign;
        }
   }
}