package clarion.options;
import clarion.system.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to represent a generic parameter in an option item of an CLAIRON
 * configuration option.
 * <li> summary of the important methods <br>
 * <ol>
 * <li> public Parameter(int majorType, int minorType, String label, String value):
 * constructor.
 * <li> public Parameter(int majorType, int minorType, String label, String[] vals):
 * constructor.
 * <li> public Parameter(int majorType, int minorType, String label, int num, String[] vals):
 * constructor.
 * <li> public Object clone(): the routine of clone this parameter.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is to represent a generic parameter in an option item of an CLAIRON
 * configuration option.
 * <li> Usage of this class. <br>
 * It is instantiated inside any GUI object where the representation of a generic
 * parameter is required.
 * </ul>
 * @author Xi Zhang
 *
 */

public class Parameter implements Cloneable {

  /** the number of values in this parameter. */
  protected int curNum;
  /** major type. 0: primitive, 1: array.  */
  protected int majorType;
  /** minor type. 0: boolean, 1: int, 2: double, 3: string. */
  protected int minorType;
  /** the parameter label. */
  protected String label;
  /** the value of the parameter. */
  protected String value;
  /** the values of the parameter. */
  protected String values[];

  /** constructor.
   *  @param majorType the major type.
   *  @param minorType the minor type.
   *  @param label the paramter label.
   *  @param value the parameter value. */
  public Parameter( int majorType, int minorType, String label, String value )
  {
     this.majorType = majorType;
     this.minorType = minorType;
     this.label = label;
     this.value = value;
  }

  /** constructor.
   *  @param majorType the major type.
   *  @param minorType the minor type.
   *  @param label the paramter label.
   *  @param values the parameter values. */
  public Parameter( int majorType, int minorType, String label, String[] vals )
  {
     this.majorType = majorType;
     this.minorType = minorType;
     this.label = label;
     this.values = new String[ vals.length ];
     curNum = vals.length;
     for( int i = 0; i < curNum; i++ )
          values[i] = new String( vals[i] );
  }

  /** constructor.
   *  @param majorType the major type.
   *  @param minorType the minor type.
   *  @param label the paramter label.
   *  @param num the number of parameter values.
   *  @param values the parameter values. */
  public Parameter( int majorType, int minorType, String label, int num, String[] vals )
  {
     this.majorType = majorType;
     this.minorType = minorType;
     this.label = label;
     this.values = new String[ num ];
     curNum = num;
     for( int i = 0; i < curNum; i++ )
          values[i] = new String( vals[i] );
  }

  /** Returns the major type of this parameter.
   *  @return the major type. */
  public int getMajorType()
  {
     return majorType;
  }

  /** Returns the minor type of this parameter.
   *  @return the minor type. */
  public int getMinorType()
  {
     return minorType;
  }

  /** Returns the label of this parameter.
   *  @return the label of this parameter. */
  public String getLabel()
  {
     return label;
  }

  /** Returns the number of parameter values.
   *  @return the number of parameter values. */
  public int getCount()
  {
     return curNum;
  }

  /** Returns the boolean value if this parameter is a boolean type.
   *  @return the boolean value. */
  public boolean getBoolValue()
  {
     return (Boolean.valueOf( value )).equals( Boolean.TRUE );
  }

  /** Returns the integer value if this parameter is an integer type.
   *  @return the integer value. */
  public int getIntValue()
  {
     return Integer.parseInt( value );
  }

  /** Returns the double value if this parameter is a double type.
   *  @return the double value. */
  public double getDoubleValue()
  {
     return Double.parseDouble( value );
  }

  /** Returns the string value if this parameter is a string type.
   *  @return the string value. */
  public String getStringValue()
  {
     return value;
  }

  /** Returns the boolean values if this parameter is a boolean array type.
   *  @return the boolean values. */
  public boolean[] getBoolValues()
  {
     boolean[] array = new boolean[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = Boolean.valueOf( values[i] ).equals( Boolean.TRUE );
     return array;
  }

  /** Returns the integer values if this parameter is an integer array type.
   *  @return the integer values. */
  public short[] getIntValues()
  {
     short[] array = new short[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = Short.parseShort( values[i] );
     return array;
  }

  /** Returns the double values if this parameter is a double array type.
   *  @return the double values. */
  public double[] getDoubleValues()
  {
     double[] array = new double[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = Double.parseDouble( values[i] );
     return array;
  }

  /** Returns the string values if this parameter is a string array type.
   *  @return the string values. */
  public String[] getStringValues()
  {
     String[] array = new String[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = values[i];
     return array;
  }

  /** Sets the parameter label.
   *  @param str the label to set. */
  public void setLabel( String str )
  {
     label = str;
  }

  /** Sets the number of paramter values.
   *  @param num the number to set. */
  public void setCount( int num )
  {
     if( num <= curNum ){
         curNum = num;
         return;
     }
     final String defaultValues[] = { "false", "1", "1.0", "val" };
     String[] temp = new String[ curNum ];
     for( int i = 0; i < curNum; i++ )
          temp[i] = new String( values[i] );
     values = new String[ num ];
     for( int i = 0; i < num; i++ )
          if( i < curNum )
              values[i] = new String( temp[i] );
          else{
              String postfix = (minorType != Global.STRING) ? "" : String.valueOf(i);
              values[i] = new String ( defaultValues[minorType] + postfix );
          }
     curNum = num;
  }

  /** Sets the boolean value if this paramter is a boolean type.
   *  @param val the value to set. */
  public void setBoolValue( boolean val )
  {
     value = String.valueOf( val );
  }

  /** Sets the integer value if this paramter is an integer type.
   *  @param val the value to set. */
  public void setIntValue( int val )
  {
     value = String.valueOf( val );
  }

  /** Sets the double value if this paramter is a double type.
   *  @param val the value to set. */
  public void setDoubleValue( double val )
  {
     value = String.valueOf( val );
  }

  /** Sets the string value if this paramter is a string type.
   *  @param val the value to set. */
  public void setStringValue( String val )
  {
     value = val;
  }

  /** Sets a specific element in the parameter array with a specific boolean value
   *  if this paramter is a boolean array type.
   *  @param idx the index of the element.
   *  @param val the value to set. */
  public void setBoolValue( int idx, boolean val )
  {
     values[idx] = String.valueOf( val );
  }

  /** Sets a specific element in the parameter array with a specific integer value
   *  if this paramter is an integer array type.
   *  @param idx the index of the element.
   *  @param val the value to set. */
  public void setIntValue( int idx, int val )
  {
     values[idx] = String.valueOf( val );
  }

  /** Sets a specific element in the parameter array with a specific double value
   *  if this paramter is a double array type.
   *  @param idx the index of the element.
   *  @param val the value to set. */
  public void setDoubleValue( int idx, double val )
  {
     values[idx] = String.valueOf( val );
  }

  /** Sets a specific element in the parameter array with a specific string value
   *  if this paramter is a string array type.
   *  @param idx the index of the element.
   *  @param val the value to set. */
  public void setStringValue( int idx, String val )
  {
     values[idx] = new String( val );
  }

  /*
  public void setBoolValues( boolean[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = String.valueOf( vals[i] );
  }

  public void setIntValues( short[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = String.valueOf( vals[i] );
  }

  public void setDoubleValues( double[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = String.valueOf( vals[i] );
  }

  public void setStringValues( String[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = vals[i];
  }
  */

  /** methods from class <em>Object</em> */
  public String toString()
  {
     String str = "Parameter major type : " + majorType + "\n";
     str += "Parameter minor type : " + minorType + "\n";
     str += "Parameter label : " + label + "\n";
     if( majorType == 0 )
         str += "Parameter value : " + value + "\n";
     else{
         str += "Parameter value : ";
         for( int i = 0; i < curNum; i++ )
              if( i < curNum - 1 )
                  str += values[i] + ",";
              else
                  str += values[i];
         str += "\n";
     }
     return str;
  }

  /** methods from class <em>Object</em> */
  public Object clone() //throws CloneNotSupportedException
  {
      Parameter param;
      if( majorType == 0 )
          param = new Parameter( majorType, minorType, label, value );
      else{
          //System.out.println( "in clone, curNum = " + curNum );
          param = new Parameter( majorType, minorType, label, curNum, values );
      }
      return param;
  }
}