package clarion.options.transfer;
import javax.swing.*;
import java.io.*;
import java.util.*;
import clarion.system.*;
import clarion.options.*;
import clarion.options.transfer.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is the template class to describe the process of transferring
*  configuration data of one susystem between GUI and <tt>Global</tt> object or
*  persistent storage.
 * <li> important methods <br>
 * <ol>
 * <li> public void load(): implement the process of loading options from persistent
 * storage.
 * <li> public void save(): implement the process of save the options to persistent
 * storage or to the associated <tt>Global</tt> object.
 * <li> public Option[] loadOptions( String path ): load options from persistent
 * storage with particulary path.
 * <li> public void saveOptions( String path, Option[] opts ): save the options to
 * persistent storage with particular path.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model. <br>
 * This class implemnts the process of loading and saving CLARION or task-specific
 * configurations for one subsystem in CLARION model or in task-specific.
 * <li> Usage of this class. <br>
 * Each <em>Transfer</em> object is associated with one CLARION or task-specific
 * subsystem. This class is to represent a generic data transfer class and the
 * concrete data transfer classes should extend this class to implement their
 * own transfer process.
 * </ul>
 * @author Xi Zhang
 *
 **/

public class Transfer {

  /** constants indicates save options to associated <tt>Global</tt> object or
   * persistent storage */
  protected final int SAVE_TO_GLOBAL = Global.SAVE_TO_GLOBAL, SAVE_TO_FILE = Global.SAVE_TO_FILES;
  /** for generic input/output format. */
  protected final int WM = 0, GKS = 1, AMN = 2, INPUT = 0, OUTPUT = 1;
  protected final int ON = 0, OFF = 1;
  protected final int TL = Global.TL, BL = Global.BL;
  protected TransferManager manager;
  /** Global settings. */
  protected Global global;
 // save to global or file
  protected int saveCode;
  /** the names of files for loading or saving the user option-selecting data. */
  protected String loadedFileName, savedFileName;
  /** the files for loading or saving the user option-selecting data. */
  protected RandomAccessFile loadedFile, savedFile;
  /** used to store the user and system default options. */
  protected Option userOptions[][][][][][], sysOptions[][][][][][];
  /** variables for locating files. */
  protected static String sysPath = Global.getSettingDir() + "\\DEFAULT";
  protected static File sysDir = new File( sysPath );
  protected String sysSettingPath = sysPath + "\\subjectGroup0";
  protected File sysSettingDir = new File( sysSettingPath );
  protected String subsysPath;

  /** constructor. instantiated with the associated <tt>TransferManager</tt> object
   * and <tt>Global</tt> object. */
  public Transfer( TransferManager manager, Global global )
  {
     this.manager = manager;
     this.global = global;
  }

  /** Sets the associated <tt>Global</tt> object.
   *  @param global the <tt>Global</tt> object to set. */
  public void setGlobal( Global global )
  {
     this.global = global;
  }

  /** The process of loading options. */
  public void load() {}
  /** The process of saving options. */
  public void save() {}
  /** the process of update. This process is necessary because some options or
   *  Global variables cannot be configured correctly at first time due to lack of
   *  information which can be known lately.  */
  public void update() {}

  /** Get an array of particular options.
   *  arguments <tt>dim1</tt> through <tt>dim5</tt> are all indices on a 6-dimension
   * array which stores all options of a particular subsystem. */
  public Option[] getOptions( int dim1, int dim2, int dim3, int dim4, int dim5 )
  {
     return userOptions[dim1][dim2][dim3][dim4][dim5];
  }

  /** Get a particular option.
   * arguments <tt>dim1</tt> through <tt>dim6</tt> are all indices on a 6-dimension
   * array which stores all options of a particular subsystem. */
  public Option getOptions( int dim1, int dim2, int dim3, int dim4, int dim5, int dim6 )
  {
     return userOptions[dim1][dim2][dim3][dim4][dim5][dim6];
  }

  /** Set the mode of saving options:
   * 0: save to <tt>Global</tt> object; 1: save to persistent storage.
   * @param code the mode of saving options. */
  public void setSaveCode( int code )
  {
     saveCode = code;
  }

  /** load options from persistent storage with particulary path. */
  public Option[] loadOptions( String path )
  {
     try {
           loadedFile = new RandomAccessFile( path + loadedFileName, "rw" );
     }
     catch ( IOException e ) {
             System.out.println( path + loadedFileName );
             JOptionPane.showMessageDialog( null, "File Operation Error in loading",
                  "Invalid File Name", JOptionPane.ERROR_MESSAGE );
     }
     String numMark = "The number of options : ";
     String[] optionMarks = {
        "Option label : ",  "Option choice number : ", "Default choice : ", "User choice : "
     };
     String[] paramMarks = {
        "Parameter number : ", "Parameter major type : ", "Parameter minor type : ",
        "Parameter label : ", "Parameter value : "
     };
     Option[] opts = null;
     try{
       loadedFile.seek( 0 );
       // option number.
       String str = loadedFile.readLine();
       int optionNum = Integer.parseInt(str.substring(numMark.length(),str.length()).trim());
       opts = new Option[ optionNum ];

       //System.out.println( "before load options, option num = " + optionNum );

       for( int i = 0; i < optionNum; i++ ){
            // option label.
            str = loadedFile.readLine();
            String optLabel = str.substring(optionMarks[0].length(),str.length());

            //System.out.println( "in loadOptions, option label = " + optLabel );

            // choice number.
            str = loadedFile.readLine();
            int choiceNum = Integer.parseInt( str.substring(optionMarks[1].length(),
                                          str.length()).trim());
            // default choice.
            str = loadedFile.readLine();
            int defaultChoice = Integer.parseInt( str.substring(optionMarks[2].length(),
                                                  str.length()).trim());
            // user choice.
            str = loadedFile.readLine();
            int subjectGroupChoice = Integer.parseInt( str.substring(optionMarks[3].length(),
                                               str.length()).trim());
            opts[i] = new Option( optLabel, choiceNum, defaultChoice, subjectGroupChoice );
            // option items.
            OptionItem[] optItems = new OptionItem[ choiceNum ];
            for( int j = 0; j < choiceNum; j++ ){
                 // item label.
                 String itemLabel = loadedFile.readLine().substring( "Item label : ".length() );
                 // item parameter number.
                 str = loadedFile.readLine();
                 int paramNum = Integer.parseInt( str.substring( paramMarks[0].length(),
                                                  str.length()).trim());
                 Parameter[] optParams = new Parameter[ paramNum ];
                 // item parameters.
                 for( int k = 0; k < paramNum; k++ ){
                      str = loadedFile.readLine();
                      int paraMajorType = Integer.parseInt(
                          str.substring(paramMarks[1].length(),str.length()).trim() );
                      str = loadedFile.readLine();
                      int paraMinorType = Integer.parseInt(
                          str.substring(paramMarks[2].length(),str.length()).trim() );
                      str = loadedFile.readLine();
                      String paramName = str.substring( paramMarks[3].length(),
                                         str.length());
                      str = loadedFile.readLine();
                      if( paraMajorType == Global.PRIM_TYPE ){
                          String paramValue = str.substring(
                                 paramMarks[4].length(), str.length() );
                          optParams[k] = new Parameter( paraMajorType, paraMinorType,
                                             paramName, paramValue );
                      }
                      else{
                          String line = str.substring(
                                 paramMarks[4].length(),str.length());
                          StringTokenizer st = new StringTokenizer( line, Global.VALUE_DELIM );
                          int count = 0;
                          String[] tokens = new String[ line.length() ];
                          while( st.hasMoreTokens() )
                                 tokens[count++] = st.nextToken();
                          String[] paramValues = new String[ count ];
                          for( int m = 0; m < count; m++ )
                               paramValues[ m ] = tokens[ m ];

                          //System.out.println( "len = " + paramValues.length );

                          optParams[k] = new Parameter( paraMajorType, paraMinorType,
                                             paramName, paramValues );
                     }
                 }
                 optItems[j] = new OptionItem( itemLabel, optParams );
            }
            opts[i].setOptionItems( optItems );
       }
     }
     catch( Exception e ){
       System.out.println( e );
     }

     //System.out.println( "in loadOptions, option length = " + opts.length );

     return opts;
  }

  /** save the options to persistent storage with particular path. */
  public void saveOptions( String path, Option[] opts )
  {
     try {
          //System.out.println( path + savedFileName );
          savedFile = new RandomAccessFile( path + savedFileName, "rw" );
          savedFile.seek( 0 );
          savedFile.writeBytes("The number of options : " + opts.length + "\n" );
          for( int i = 0; i < opts.length; i++ ){
               //System.out.println( opts[i].toString() );
               savedFile.writeBytes( opts[i].toString() );
          }
     }
     catch ( IOException e ) {
             JOptionPane.showMessageDialog( null, "File Operation Error in writing!",
                  "Invalid File Name", JOptionPane.ERROR_MESSAGE );
     }
  }

  /** Sets up the inpt/output dimension information for WM, GKS, AMN. */
  public void setupGenericFormat( int compType, int inOut, int netIdx, Option opt )
  {
     int itemNum = (compType == WM) ? global.getWmSize() : 1;
     boolean escFlag = global.isESChunkOn();
     boolean allowSysFlag = opt.getOptionItem(0).getParameter(0).getBoolValue();
     boolean genericFlag = escFlag && allowSysFlag;
     int compNum = genericFlag ? Global.GENERIC_COMP_NUM : 2;
     int[][] dimNums = new int[itemNum][compNum];
     int dimNum = 0;
     for( int i = 0; i < itemNum; i++ ){
          if( genericFlag ){
             // EX-input
             dimNums[i][Global.GENERIC_EX_INPUT] = global.getFeatureNums()[Global.FT_INPUT];
             // GS-input
             dimNums[i][Global.GENERIC_GS_INPUT] = (global.getGsOption()!=Global.GS_NON)?global.getFeatureNums()[Global.FT_GS]:0;
             // EX-action
             dimNums[i][Global.GENERIC_EX_ACT] = global.getFeatureNums()[Global.FT_OUTPUT];
             // GS-action
             dimNums[i][Global.GENERIC_GS_ACT] = (global.getGsOption()!=Global.GS_NON)?global.getFeatureNums()[Global.FT_GS]+1:0;
             // WM-action
             dimNums[i][Global.GENERIC_WM_ACT] = Global.ACT_DIM_NUM_WM;
             // NACS control
             dimNums[i][Global.GENERIC_NACS_ACT] = Global.ACT_DIM_NUM_NACS_CONTROL;
             // WM flags
             dimNums[i][Global.GENERIC_WM_FLAG] = (global.getWmFlagNum()==0)?0:1;
             // user-specified.
             dimNums[i][Global.GENERIC_USER_SPEC] = 0;
             if( opt.getOptionItem(0).getParameter(1).getIntValues()[0] != Global.EMPTY )
                 dimNums[i][Global.GENERIC_USER_SPEC] = opt.getOptionItem(0).getParameter(1).getCount();
         }
         else{
             // WM flags
             dimNums[i][0] = (global.getWmFlagNum()==0)?0:1;
             // user-specified.
             dimNums[i][1] = 0;
             if( opt.getOptionItem(0).getParameter(1).getIntValues()[0] != Global.EMPTY )
                 dimNums[i][1] = opt.getOptionItem(0).getParameter(2).getCount();
         }
         for( int j = 0; j < compNum; j++ )
              dimNum += dimNums[i][j];
     }
     // intializes variables
     short[] dimTypes = new short[ dimNum ];
     short[] dimIdxs = new short[ dimNum ];
     short[] dimVals = new short[ dimNum ];
     short[][] featInfo = new short[ dimNum ][2];
     // fill in dimTypes and dimIdxs
     short[] itemDimTypes = { Global.FT_INPUT, Global.FT_GS, Global.FT_OUTPUT, Global.FT_SYS,
                              Global.FT_SYS, Global.FT_SYS, Global.FT_SYS, 0 };
     short[] itemDimFeatStarts = { 0, 0, 0, Global.FT_SYS_GS_ACT, Global.FT_SYS_WM_ACT,
                                   Global.FT_SYS_NACS_ACT, Global.FT_SYS_WM_FLAG, 0  };
     short dimOffset = 0;
     for( int i = 0; i < itemNum; i++ )
       for( int j = 0; j < compNum; j++ ){
         if( genericFlag ){
             if( j != Global.GENERIC_GS_ACT || j!= Global.GENERIC_USER_SPEC ){
                 for( int k = 0; k < dimNums[i][j]; k++ ){
                      dimTypes[dimOffset + k] = itemDimTypes[j];
                      dimIdxs[dimOffset + k] = (short)(itemDimFeatStarts[j] + k);
                 }
             }
             if( j == Global.GENERIC_GS_ACT ){ // GS action: action dim + para dims + goal-ID dim.
                 dimTypes[dimOffset] = itemDimTypes[j];
                 dimIdxs[dimOffset] = itemDimFeatStarts[j];
                 for( int k = 1; k < dimNums[i][j]; k++ ){
                      dimTypes[dimOffset + k] = Global.FT_GS;
                      dimIdxs[dimOffset + k] = (short)(k - 1);
                 }
             }
             if( j == Global.GENERIC_USER_SPEC ){
                 for( int k = 0; k < dimNums[i][j]; k++ ){
                      dimTypes[dimOffset + k] = opt.getOptionItem(0).getParameter(1).getIntValues()[k];
                      dimIdxs[dimOffset + k] = opt.getOptionItem(0).getParameter(2).getIntValues()[k];
                 }
             }
         }
         else{
             if( j == 0 ){ // WM flag dimension.
                 for( int k = 0; k < dimNums[i][j]; k++ ){
                      dimTypes[dimOffset + k] = Global.FT_SYS;
                      dimIdxs[dimOffset + k] = (short)(Global.FT_SYS_WM_FLAG);
                 }
             }
             if( j == 1 ){ // user-defined dimensions
                 for( int k = 0; k < dimNums[i][j]; k++ ){
                      int end = Math.min( opt.getOptionItem(0).getParameter(1).getCount()-1, k );
                      dimTypes[dimOffset + k] = opt.getOptionItem(0).getParameter(1).getIntValues()[end];
                      dimIdxs[dimOffset + k] = opt.getOptionItem(0).getParameter(2).getIntValues()[k];
                 }
             }
         }
         dimOffset += dimNums[i][j];
      }
      for( int i = 0; i < dimNum; i++ ){
           dimVals[i] = (short)global.getFeatureLength(dimTypes[i], dimIdxs[i]);
           featInfo[i][0] = dimTypes[i];
           featInfo[i][1] = dimIdxs[i];
      }
      // initialize the relevant global variables.
      switch( compType ){
          case WM :
               global.setIdnInputDimVals( netIdx, dimVals );
               global.setIdnInputFeatureInfo( netIdx, featInfo );
               break;
          case GKS :

               //System.out.println( "feature num = " + featInfo.length );
               //System.out.println( "feature 0 info = " + featInfo[0][0] );

               if( inOut == INPUT ){
                   global.setGksInputDimVals( dimVals );
                   global.setGksInputFeatureInfo( featInfo );
               }
               else{
                   global.setGksOutputDimVals( dimVals );
                   global.setGksOutputFeatureInfo( featInfo );
               }
               break;
          case AMN :
               if( inOut == INPUT ){
                   global.setAmnInputDimVals( netIdx, dimVals );
                   global.setAmnInputFeatureInfo( netIdx, featInfo );
               }
               else{
                   global.setAmnOutputDimVals( netIdx, dimVals );
                   global.setAmnOutputFeatureInfo( netIdx, featInfo );
               }
               break;
      }
 }

  /** search a most suitable setting directory for a particular subystem.
   * @param subSysName the name of the subsystem. */
  public String searchFile( String subSysName )
  {
     String path = global.getSettingDir() + "\\" + global.getTaskName() + "\\" +
                   global.getTaskSettingName() + "\\" + subSysName;
     File file = new File( path );
     if( file.exists() )
         return path + "\\";
     // search a most suitable setting file.
     String taskPath = global.getSettingDir() + "\\" + global.getTaskName();
     String taskSettingPath = taskPath + "\\" + global.getTaskSettingName();
     String[] settingNames = manager.searchCurrentSettings();
     boolean existFlag = false;
     for( int i = 0; i < settingNames.length; i++ ){
          String existSubsysSettingPath = taskPath + "\\" + settingNames[i] + "\\" + subSysName;
          File existSubsysSettingDir = new File( existSubsysSettingPath );
          if( existSubsysSettingDir.exists() ){
              path = existSubsysSettingPath;
              existFlag = true;
              break;
          }
     }
     if( !existFlag ){
         String defaultSubsysSettingPath = sysSettingPath + "\\" + subSysName;
         path = defaultSubsysSettingPath;
     }
     return path + "\\";
  }

  /** search a most suitable setting directory for a particular subystem name
   *  and file name for a component in that subsystem.
   * @param subSysName the name of the subsystem.
   * @param fileName file name for component. */
  public String searchFile( String subSysName, String fileName )
  {
     String path = subsysPath;
     File file = new File( subsysPath + fileName );
     if( file.exists() )
         return path;
     // search a most suitable setting file.
     String taskPath = global.getSettingDir() + "\\" + global.getTaskName();
     String taskSettingPath = taskPath + "\\" + global.getTaskSettingName();
     String[] settingNames = manager.searchCurrentSettings();
     boolean existFlag = false;
     for( int i = 0; i < settingNames.length; i++ ){
          String existSubsysSettingPath = taskPath + "\\" + settingNames[i] + "\\" + subSysName;
          File existSubsysSettingDir = new File( existSubsysSettingPath );
          if( existSubsysSettingDir.exists() ){
              File existAemFile = new File( existSubsysSettingPath + "\\" + fileName );
              if( existAemFile.exists() ){
                  path = existSubsysSettingPath + "\\";
                  existFlag = true;
                  break;
              }
          }
     }
     if( !existFlag ){
         String defaultSubsysSettingPath = sysSettingPath + "\\" + subSysName + "\\";
         path = defaultSubsysSettingPath;
     }
     return path;
  }
}
