package tasksimulators.newsimu.PC;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.net.auxiliary.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class extends the <tt>TaskClarion</tt> class and implements all
 * of the necessary methods specific to PC task.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public short[][][] getEligibility( int netIdx ) :  eligibility of ACS networks,
 * return the eligible condition of the ACS network.
 * <li> public IrlRuleForm getIrlRule( int groupIdx, int netIdx, int setIdx, int ruleIdx ):
 * return the IRL rule for action decision.
 * <li> public short[][] getFrCondition( int groupIdx, int netIdx, int ruleIdx ):
 * return the  conditions. of the specified Fixed Rule
 * <li> public short[][][] getFrAction( int groupIdx, int netIdx, int ruleIdx ) :
 * return the  actions of the specified Fixed Rule.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * It is a part of the PC task simulator.
 * <li> Usage of this class. <br>
 * This class is instantiated inside a <tt>TaskAgent</tt> object.
 * </ul>
 * @author Xi Zhang
 */

public class TaskClarionPC extends TaskClarion {

  /** names of subject groups. */
  final int CONTROL_GROUP = 0, ORIGINAL_GROUP = 1;
  final int MEMORY_GROUP = 2, SIMPLE_GROUP = 3;
  /** variables for I/O dimension info. */
  final int INPUT_DIM_NUM = 1, ACTION_NUM = 12, DIM_VAL_NUM = 12, TARGET = 5;
  final int IRL_SET_NUM = 4;
  /** variables for contructing IRL rule set. */
  Vector[] irlRules;
  int[] a = { 1, 2 }; // a's b's and c's
  int[] b = { -2, -1, 0, 1, 2 }; // values to try in the rules
  int[] c = { -2,- 1, 1, 2 };
  /** variables for fixed rules. */
  short[] frConditions[][][];
  short[] frActions[][][][];
  { // initialize FR rules.
    frConditions = new short[ 1 ][ ACTION_NUM ][ INPUT_DIM_NUM ][ DIM_VAL_NUM ];
    for( int i = 0; i < 1; i++ )
      for( int j = 0; j < ACTION_NUM; j++ ){
           for( int k = 0; k < INPUT_DIM_NUM; k++ ){
                    for( int m = 0; m < DIM_VAL_NUM; m++ )
                         frConditions[i][j][k][m] = 0;
                    frConditions[i][j][k][j] = 1;
                }
      }
    frActions = new short[ 1 ][ ACTION_NUM ][2][1][1];
    for( int i = 0; i < ACTION_NUM; i++ ) {
         int n = 0, iRand = ( int ) ( Math.random() * 3 );
         switch( iRand ) {
         case 0:  n = 1;
                  break;
         case 1:  n = 0;
                  break;
         case 2:  n = -1;
                  break;
         }
        frActions[0][i][Global.ACTION_NORMAL][0][0] = (short)(( TARGET + i + n ) / 2 );
    }
  }

  /** constructor to construct the IRL sets. */
  public TaskClarionPC()
  {
     generateIrlRules();
  }

  /** Get the eligibility condition of a ACS network. See <tt>TaskClarion</tt> on
   * the details of the eligibility format.
   * @param netIdx the ACS network index.
   * @return the eligible condition of the network. */
  public short[][][] getEligibility( int netIdx )
  {
    short[][][] eligCondition = {
                   {
                      {0},            // relevant dims: 0 in this condition.
                      {1,1,1,1,1},    // all values are eligible.
                   },
                };
     return eligCondition;
  }

  /** Get the IRL rule for action decision making.
   * @param groupIdx the subject group index.
   * @param netIdx the ACS network index.
     @param setIdx  the rule set index.
     @param ruleIdx  the rule index on that rule set.
     @return the IRL rule. */
  public IrlRuleForm getIrlRule( int groupIdx, int netIdx, int setIdx, int ruleIdx )
  {
     return (IrlRuleForm)(irlRules[setIdx].elementAt(ruleIdx));
  }

  /** Get the condition of the specific Fixed Rule for action decision making.
   * @param groupIdx the subject group index.
   * @param netIdx the ACS network index.
     @param ruleIdx the rule index on that rule set.
     @return the the condition of the specific Fixed Rule. */
  public short[][] getFrCondition( int groupIdx, int netIdx, int ruleIdx )
  {
     if( groupIdx == MEMORY_GROUP )
         return frConditions[netIdx][ruleIdx];
     else
         return (short[][]) null;
  }

  /** Get the action of the specific Fixed Rule for action decision making.
   * @param groupIdx the subject group index.
   * @param netIdx the ACS network index.
     @param ruleIdx the rule index on that rule set.
     @return the the action of the specific Fixed Rule. */
  public short[][][] getFrAction( int groupIdx, int netIdx, int ruleIdx )
  {
     if( groupIdx == MEMORY_GROUP )
         return frActions[netIdx][ruleIdx];
     else
         return (short[][][])null;
  }

  private void generateIrlRules()
  {
     irlRules = new Vector[IRL_SET_NUM];
     for( int generation = 1; generation < 5; generation++ ){
        int ruleSetIdx = -1;
        int[] tempB = new int[5];
        int[] tempC;
        if( ( generation % 2 ) == 0 )
            tempC = new int[4];
        else
            tempC = new int[1];
        switch( generation ){
           case 1:
                ruleSetIdx = 0;
                tempB = b;
                tempC[0] = 0;
                break;
           case 2:
                ruleSetIdx = 1;
                tempB = b;
                tempC = c;
                break;
           case 3:
                ruleSetIdx = 2;
                tempB = b;
                tempC[0] = 0;
                 break;
           case 4:
                 ruleSetIdx = 3;
                 tempB = b;
                 tempC = c;
                 break;
        }
        // generate a new rule set.
        irlRules[generation-1] = new Vector();
        for( int i = 0; i < a.length; i++ )
          for( int j = 0; j < tempB.length; j++ )
             for( int k = 0; k < tempC.length; k++ )
                  irlRules[generation-1].addElement( new IrlRuleFormPC( ruleSetIdx,
                                         a[i], tempB[j], tempC[k], TARGET, 0 ) );
    }
  }
}
